from bicycleparameters import io

def test_filename_to_dict():
    out = io.filename_to_dict('BrowserFrameCompoundFirst1.mat')
    ans = {'angleOrder': 'First',
           'bicycle': 'Browser',
           'part': 'Frame',
           'pendulum': 'Compound',
           'trial': '1'}
    assert out == ans



#!/usr/bin/env python

import numpy as np
import matplotlib.pyplot as plt


def compare_bode_bicycles(bikes, speed, u, y, fig=None):
    """Returns a figure with the Bode plots of multiple bicycles.

    Parameters
    ----------
    bikes : list
        A list of bicycleparameters.Bicycle instances.
    speed : float
        The speed at which to evaluate the system.
    u : integer
        An integer between 0 and 1 corresponding to the inputs roll torque
        and steer torque.
    y : integer
        An integer between 0 and 3 corresponding to the inputs roll rate,
        steer rate, roll angle and steer angle.

    Returns
    -------
    fig : matplotlib.Figure instance
        The Bode plot.

    Notes
    -----
    The phases are matched around zero degrees at with respect to the first
    frequency.

    """

    if fig is None:
        fig = plt.figure()

    for bike in bikes:
        bike.plot_bode(speed, u, y, label=bike.bicycleName, fig=fig)

    # take care of phase misalignment
    phaseLines = fig.ax2.lines
    for line in phaseLines:
        firstValue = line.get_ydata()[0]
        n = np.ceil(np.floor(abs(firstValue / 180.)) / 2.)
        line.set_ydata(line.get_ydata() - np.sign(firstValue) * n * 360.)

    return fig


def plot_eigenvalues(bikes, speeds, colors=None, linestyles=None,
                     largest=False, show=False):
    '''Returns a figure with the eigenvalues vs speed for multiple bicycles.

    Parameters
    ----------
    bikes : list
        A list of Bicycle objects.
    speeds : ndarray, shape(n,)
        An array of speeds.
    colors : list
        A list of matplotlib colors for each bicycle.
    linestyles : list
        A list of matplotlib linestyles for each bicycle.
    largest : boolean
        If true, only plots the largest eigenvalue.

    Returns
    -------
    fig : matplotlib figure

    '''

    numBikes = len(bikes)

    if not colors:
        # define some colors for the parts
        cmap = plt.get_cmap('gist_rainbow')
        colors = []
        for i in range(numBikes):
            colors.append(cmap(1. * i / numBikes))

    if not linestyles:
        linestyles = ['-'] * numBikes

    for i, bike in enumerate(bikes):
        if i == 0:
            fig = None
        fig = bike.plot_eigenvalues_vs_speed(speeds, fig=fig, color=colors[i],
                                             largest=largest, generic=True,
                                             linestyle=linestyles[i],
                                             show=False)
        plt.legend()
        plt.axis('tight')

    if show is True:
        plt.show()

    return fig


def _generate_ellipse_plot_data(x_center, y_center, ax1, ax2, a, b, N):
    # x_center, y_center the coordinates of ellipse center
    # ax1 ax2 two orthonormal vectors representing the ellipse axis directions
    # a, b the ellipse parameters
    t = np.linspace(0, 2 * np.pi, N)
    # ellipse parameterization with respect to a system of axes of directions
    # a1, a2
    xs = a * np.cos(t)
    ys = b * np.sin(t)
    # rotation matrix
    R = np.array([ax1, ax2]).T
    # coordinate of the ellipse points with respect to the system of axes[1,
    # 0], [0, 1] with origin(0, 0)
    xp, yp = np.dot(R, [xs, ys])
    x = xp + x_center
    y = yp + y_center
    return x, y



#!/usr/bin/env python

from math import pi
import numpy as np
from uncertainties import unumpy, umath

from .bicycle import trail, lambda_from_abc


def calculate_benchmark_geometry(mp, par):
    '''Returns the wheelbase, steer axis tilt and the trail.

    Parameters
    ----------
    mp : dictionary
        Dictionary with the measured parameters.
    par : dictionary
        Dictionary with the benchmark parameters.

    Returns
    -------
    par : dictionary
        par with the benchmark geometry added.

    '''
    # calculate the wheel radii
    par['rF'] = mp['dF'] / 2. / pi / mp['nF']
    par['rR'] = mp['dR'] / 2. / pi / mp['nR']

    # calculate the frame/fork fundamental geometry
    if 'w' in mp.keys():  # if there is a wheelbase
        # steer axis tilt in radians
        par['lam'] = pi / 180. * (90. - mp['gamma'])
        # wheelbase
        par['w'] = mp['w']
        # fork offset
        forkOffset = mp['f']
    else:
        h = (mp['h1'], mp['h2'], mp['h3'], mp['h4'], mp['h5'])
        d = (mp['d1'], mp['d2'], mp['d3'], mp['d4'], mp['d'])
        a, b, c = calculate_abc_geometry(h, d)
        par['lam'] = lambda_from_abc(par['rF'], par['rR'], a, b, c)
        par['w'] = (a + b) * umath.cos(par['lam']) + c * umath.sin(par['lam'])
        forkOffset = b

    # trail
    par['c'] = trail(par['rF'], par['lam'], forkOffset)[0]

    # this will simply transfer the lateral force point through to the
    # benchmark parameters, as they are in the benchmark coordinate system
    try:
        par['xcl'] = mp['xcl']
        par['zcl'] = mp['zcl']
    except KeyError:
        pass

    return par


def calculate_abc_geometry(h, d):
    '''Returns the perpendicular distance geometry for the bicycle from the raw
    measurements.

    Parameters
    ----------
    h : tuple
        Tuple containing the measured parameters h1-h5.
        (h1, h2, h3, h4, h5)
    d : tuple
        Tuple containing the measured parameters d1-d4 and d.
        (d1, d2, d3, d4, d)

    Returns
    -------
    a : ufloat or float
        The rear frame offset.
    b : ufloat or float
        The fork offset.
    c : ufloat or float
        The steer axis distance.

    '''
    # extract the values
    h1, h2, h3, h4, h5 = h
    d1, d2, d3, d4, d = d
    # get the perpendicular distances
    a = h1 + h2 - h3 + .5 * d1 - .5 * d2
    b = h4 - .5 * d3 - h5 + .5 * d4
    c = umath.sqrt(-(a - b)**2 + (d + .5 * (d2 + d3))**2)
    return a, b, c


def calculate_l1_l2(h6, h7, d5, d6, l):
    '''Returns the distance along (l2) and perpendicular (l1) to the steer axis
    from the front wheel center to the handlebar reference point.

    Parameters
    ----------
    h6 : float
        Distance from the table to the top of the front axle.
    h7 : float
        Distance from the table to the top of the handlebar reference circle.
    d5 : float
        Diameter of the front axle.
    d6 : float
        Diameter of the handlebar reference circle.
    l : float
        Outer distance from the front axle to the handlebar reference circle.

    Returns
    -------
    l1 : float
        The distance from the front wheel center to the handlebar reference
        center perpendicular to the steer axis. The positive sense is if the
        handlebar reference point is more forward than the front wheel center
        relative to the steer axis normal.
    l2 : float
       The distance from the front wheel center to the handlebar reference
       center parallel to the steer axis. The positive sense is if the
       handlebar reference point is above the front wheel center with reference
       to the steer axis.

    '''
    r5 = d5 / 2.
    r6 = d6 / 2.
    l1 = h7 - h6 + r5 - r6
    l0 = l - r5 - r6
    gamma = umath.asin(l1 / l0)
    l2 = l0 * umath.cos(gamma)
    return l1, l2


def calc_two_link_angles(L1, L2, D):
    '''Solves a simple case of the two-link revolute joint inverse kinematics
    problem. Both output angles are positive. The simple case is that the end
    of the second link lies on the x-axis.

    Parameters
    ----------
    L1 : float
        Length of the first link.
    L2 : float
        Length of the second link.
    D : float
        Distance from the base of first link to the end of the second link.

    Returns
    -------
    theta1 : float
        (radians) Angle between x-axis and first link; always positive.
    theta2 : float
        (radians) Angle between first link and second link; always positive.

    '''

    theta1 = np.arccos((L1**2 + D**2 - L2**2) / (2.0 * L1 * D))
    theta2 = theta1 + np.arcsin(L1 / L2 * np.sin(theta1))

    return theta1, theta2


def fwheel_to_handlebar_ref(lam, l1, l2):
    '''Returns the distance along the benchmark coordinates from the front
    wheel center to the handlebar reference center.

    Parameters
    ----------
    lam : float
        Steer axis tilt.
    l1, l2 : float
        The distance from the front wheel center to the handlebar refernce
        center perpendicular to and along the steer axis.

    Returns
    -------
    u1, u2 : float

    '''

    u1 = l2 * umath.sin(lam) - l1 * umath.cos(lam)
    u2 = u1 / umath.tan(lam) + l1 / umath.sin(lam)
    return u1, u2


def fundamental_geometry_plot_data(par):
    '''Returns the coordinates for line end points of the bicycle fundamental
    geometry.

    Parameters
    ----------
    par : dictionary
        Benchmark bicycle parameters.

    Returns
    -------
    x : ndarray
    z : ndarray

    '''
    d1 = np.cos(par['lam']) * (par['c'] + par['w'] - par['rR'] *
                               np.tan(par['lam']))
    d3 = -np.cos(par['lam']) * (par['c'] - par['rF'] * np.tan(par['lam']))
    x = np.zeros(4, dtype=object)
    z = np.zeros(4, dtype=object)
    x[0] = 0.
    x[1] = d1 * np.cos(par['lam'])
    x[2] = par['w'] - d3 * np.cos(par['lam'])
    x[3] = par['w']
    z[0] = -par['rR']
    z[1] = -par['rR'] - d1 * np.sin(par['lam'])
    z[2] = -par['rF'] + d3 * np.sin(par['lam'])
    z[3] = -par['rF']

    return x, z


def distance_to_steer_axis(w, c, lam, point):
    """Returns the minimal distance from the steer axis to the given point when
    the bicycle is in the nominal configuration.

    Parameters
    ----------
    w : float or ufloat
        Wheelbase.
    c : float or ufloat
        Trail.
    lam : float or ufloat
        Steer axis tilt in radians.
    point : narray, shape(3,)
        A point that lies in the symmetry plane of the bicycle.

    Returns
    -------
    d : float or ufloat
        The minimal distance from the given point to the steer axis.

    """
    pointOnAxis1 = np.array([w + c,
                             0.,
                             0.])
    pointOnAxis2 = (pointOnAxis1 + np.array([-umath.sin(lam), 0.,
                                             -umath.cos(lam)]))
    pointsOnLine = np.array([pointOnAxis1, pointOnAxis2]).T

    # this is the distance from the assembly com to the steer axis
    return point_to_line_distance(point, pointsOnLine)


def point_to_line_distance(point, pointsOnLine):
    '''Returns the minimal distance from a point to a line in three
    dimensional space.

    Parameters
    ----------
    point : ndarray, shape(3,)
        The x, y, and z coordinates of a point.
    pointsOnLine : ndarray, shape(3,2)
        The x, y, and z coordinates of two points on a line. Rows are
        coordinates and columns are points.

    Returns
    -------
    distance : float
        The minimal distance from the line to the point.

    '''
    x1 = pointsOnLine[:, 0]
    x2 = pointsOnLine[:, 1]
    x0 = point

    def norm(v):
        return unumpy.sqrt(np.dot(v, v))

    distance = norm(np.cross((x0 - x1), (x0 - x2))) / norm(x2 - x1)

    return distance


def project_point_on_line(line, point):
    '''Returns point of projection.

    Parameters
    ----------
    line : tuple
        Slope and intercept of the line.
    point : tuple
        Location of the point.

    Returns
    -------
    newPoint : tuple
        The location of the projected point.

    '''
    m, b = line
    c, d = point
    x = (m * d + c - m * b) / (m**2. + 1.)
    y = (m**2. * d + m * c + b) / (m**2. + 1.)
    return x, y


def vec_angle(v1, v2):
    '''Returns the interior angle between two vectors using the dot product.
    Inputs do not need to be unit vectors.

    Parameters
    ----------
    v1 : np.array (3,1)
        input vector.
    v2 : np.array (3,1)
        input vector.

    Returns
    -------
    angle : float
        (radians) interior angle between v1 and v2.
    '''
    return np.arccos(float(np.dot(v1.T, v2)) / (np.linalg.norm(v1) *
                                                np.linalg.norm(v2)))


def vec_project(vec, direction):
    '''Vector projection into a plane, where the plane is defined by a
    normal vector.

    Parameters
    ----------
    vec : np.array(3,1)
        vector to be projected into a plane
    direction : int or np.array, shape(3,)
        If int, it is one of the three orthogonal directions, (0,1 or 2) of
        the input vector (essentially, that component of vec is set to zero).
        If np.array, can be in any direction (not necessarily a coordinate
        direction).

    Returns
    -------
    vec_out : np.array(3,1)
        Projected vector.

    '''
    vec = vec.flatten()
    if type(direction) == int:
        unitdir = np.zeros(3)
        unitdir[direction] = 1
    elif type(direction) == np.array:
        unitdir = direction / np.linalg.norm(direction)
    proj = vec - np.dot(vec, unitdir)
    return proj.reshape((3, 1))



#!/usr/bin/env python

import re
import os
import numpy as np
from uncertainties import ufloat_fromstr
from scipy.io import loadmat


def filename_to_dict(filename):
    '''Returns a dictionay of values based on the pendulum data file name.

    '''
    o = space_out_camel_case(os.path.splitext(filename)[0], output='list')
    # this only accounts for single digit trial numbers
    trial = o[-1][-1]
    o[-1] = o[-1][:-1]
    o.append(trial)
    breakdown = ['bicycle', 'part', 'pendulum', 'angleOrder', 'trial']
    dat = {}
    for word, val in zip(breakdown, o):
        dat[word] = val
    return dat


def load_parameter_text_file(pathToFile):
    """Returns a dictionary of float and/or ufloat parameters from a parameter
    file.

    Parameters
    ----------
    pathToFile : string
        The path to the text file with the parameters listed in the specified
        format.

    Returns
    -------
    parameters : dictionary
        A dictionary of the values stored in the text files.

    For example::

        c = 0.08 +/- 0.01
        d=0.314+/-0.002
        t = 0.1+/-0.01, 0.12+/-0.02
        whb = 0.5

    The first item on the line must be the variable name and the second is an
    equals sign. The values to the right of the equal sign much may or may not
    contain an uncertainty designated by `+/-`. Multiple comma seperated values
    will be averaged.

    """

    parameters = {}
    # parse the text file
    with open(pathToFile, 'r') as f:
        for line in f:
            # ignore lines that start with a hash
            if line[0] != '#':
                # remove any whitespace characters and comments at the end of
                # the line, then split the right and left side of the equality
                equality = line.strip().split('#')[0].split('=')
                # ['a ', ' 0.1 +/- 0.05 , 0.09 +/- 0.05']
                valList = equality[1].strip().split(',')
                # ['0.1 +/- 0.05 ', ' 0.09 +/- 0.05']
                if '+/-' in equality[1]:
                    values = [ufloat_fromstr(x) for x in valList]
                else:
                    values = [float(x) for x in valList]
                # store in dictionary
                parameters[equality[0].strip()] = np.mean(values)

    return parameters


def load_pendulum_mat_file(pathToFile):
    '''Returns a dictionay containing the data from the pendulum data mat file.

    '''
    pendDat = {}
    loadmat(pathToFile, mdict=pendDat)
    # clean up the matlab imports
    del(pendDat['__globals__'], pendDat['__header__'], pendDat['__version__'])
    for k, v in pendDat.items():
        try:
            # change to an ascii string
            pendDat[k] = v[0].encode('ascii')
        except:
            # if an array of a single number
            if np.shape(v)[0] == 1:
                pendDat[k] = v[0][0]
            # else if the notes are empty
            elif np.shape(v)[0] == 0:
                pendDat[k] = ''
            # else it is the data which needs to be a one dimensional array
            else:
                pendDat[k] = v.reshape((len(v),))
    return pendDat


def remove_uncertainties(dictionary):
    '''Returns a dictionary with the uncertainties removed.'''
    noUncert = {}
    for k, v in dictionary.items():
        try:
            # this is the case if the value is a single uncertainty
            noUncert[k] = v.nominal_value
        except AttributeError:
            # this is the case if the value is an array of ufloats
            try:
                noUncert[k] = [x.nominal_value for x in v]
            except TypeError:
                # this is the case if the value is a float
                noUncert[k] = v
    return noUncert


def write_parameter_text_file(pathToTxtFile, parDict):
    '''Writes parameter set to file.

    Parameters
    ----------
    pathToTxtFile : string
        The path to the file to write the parameters.
    pardict : dictionary
        A dictionary of parameters for the bicycle.

    Returns
    -------
    saved : boolean
        True if the file was saved and false if not.

    '''

    # make the Parameters directory if it doesn't exist
    head, tail = os.path.split(pathToTxtFile)
    if not os.path.isdir(head):
        print("Created direcotry %s" % head)
        os.makedirs(head)

    try:
        f = open(pathToTxtFile)
        f.close()
        del f
        ans = None
        while ans != 'y' and ans != 'n':
            ans = input("%s exists already. Are you sure you want"
                        " to overwrite it? (y or n)\n" % pathToTxtFile)
        if ans == 'y':
            f = open(pathToTxtFile, 'w')
    except IOError:
        f = open(pathToTxtFile, 'w')

    try:
        f
        keys = sorted(parDict.keys())
        for key in keys:
            f.write(key + ' = ' + str(parDict[key]) + '\n')
        f.close()
        print("Parameters saved to %s" % pathToTxtFile)
        return True
    except UnboundLocalError:
        print("%s was not saved." % pathToTxtFile)
        return False


def space_out_camel_case(s, output='string'):
    """Adds spaces to a camel case string. Failure to space out string returns
    the original string.

    Examples
    --------
    >>> space_out_camel_case('DMLSServicesOtherBSTextLLC')
    'DMLS Services Other BS Text LLC'
    >>> space_out_camel_case('DMLSServicesOtherBSTextLLC', output='list')
    ['DMLS', 'Services', 'Other', 'BS', 'Text', 'LLC']

    """
    if output == 'string':
        return re.sub('((?=[A-Z][a-z])|(?<=[a-z])(?=[A-Z]))', ' ', s).strip()
    elif output == 'list':
        string = re.sub('((?=[A-Z][a-z])|(?<=[a-z])(?=[A-Z]))', ' ', s).strip()
        return string.split(' ')
    else:
        raise ValueError


def write_periods_to_file(pathToRawFile, mp):
    '''Writes the provided periods to file.

    Parameters
    ----------
    pathToRawFile : string
        The path to the <bicycle name>Measured.txt file
    mp : dictionary
        The measured parameters dictionary. Should contain complete period
        data.

    '''

    # clear any period data from the file
    f = open(pathToRawFile, 'r')
    baseData = ''
    for line in f:
        if not line.startswith('T'):
            baseData += line
    f.close()
    # add the periods to the base data
    periodKeys = [x for x in mp.keys() if x.startswith('T')]
    periodKeys.sort()
    withPeriods = baseData
    for k in periodKeys:
        withPeriods += k + ' = ' + str(mp[k]) + '\n'

    # write it to the file
    f = open(pathToRawFile, 'w')
    f.write(withPeriods)
    f.close()



import numpy as np
from pytest import raises

from ..parameter_sets import Meijaard2007ParameterSet
from ..models import Meijaard2007Model

meijaard2007_parameters = {  # dictionary of the parameters in Meijaard 2007
    'IBxx': 9.2,
    'IBxz': 2.4,
    'IByy': 11.0,
    'IBzz': 2.8,
    'IFxx': 0.1405,
    'IFyy': 0.28,
    'IHxx': 0.05892,
    'IHxz': -0.00756,
    'IHyy': 0.06,
    'IHzz': 0.00708,
    'IRxx': 0.0603,
    'IRyy': 0.12,
    'c': 0.08,
    'g': 9.81,
    'lam': np.pi/10.0,
    'mB': 85.0,
    'mF': 3.0,
    'mH': 4.0,
    'mR': 2.0,
    'rF': 0.35,
    'rR': 0.3,
    'v': 5.0,
    'w': 1.02,
    'xB': 0.3,
    'xH': 0.9,
    'zB': -0.9,
    'zH': -0.7,
}


def test_parse_parameter_overrides():

    parameter_set = Meijaard2007ParameterSet(meijaard2007_parameters, True)

    model = Meijaard2007Model(parameter_set)

    par, array_keys, array_len = model._parse_parameter_overrides(zH=4.0)
    assert par['zH'] == 4.0
    assert array_keys == []
    assert array_len is None

    par, array_keys, array_len = model._parse_parameter_overrides(
        zH=4.0, IBxx=6.0)
    assert par['zH'] == 4.0
    assert par['IBxx'] == 6.0
    assert array_keys == []
    assert array_len is None

    par, array_keys, array_len = model._parse_parameter_overrides(
        zH=4.0, IBxx=[1.0, 2.0, 3.0])
    assert par['zH'] == 4.0
    assert array_keys == ['IBxx']
    np.testing.assert_allclose(par['IBxx'], [1.0, 2.0, 3.0])
    assert array_len == 3

    par, array_keys, array_len = model._parse_parameter_overrides(
        v=[1.0, 2.0, 3.0])
    np.testing.assert_allclose(par['v'], [1.0, 2.0, 3.0])
    assert array_keys == ['v']
    assert array_len == 3

    par, array_keys, array_len = model._parse_parameter_overrides(
        g=[1.0, 2.0, 3.0])
    np.testing.assert_allclose(par['g'], [1.0, 2.0, 3.0])
    assert array_keys == ['g']
    assert array_len == 3

    par, array_keys, array_len = model._parse_parameter_overrides(
        zH=4.0, v=[1.0, 2.0, 3.0])
    assert par['zH'] == 4.0
    assert array_keys == ['v']
    np.testing.assert_allclose(par['v'], [1.0, 2.0, 3.0])
    assert array_len == 3

    par, array_keys, array_len = model._parse_parameter_overrides(
        zH=[5.0, 6.0, 7.0], v=[1.0, 2.0, 3.0])
    assert array_keys == ['zH', 'v']
    np.testing.assert_allclose(par['v'], [1.0, 2.0, 3.0])
    np.testing.assert_allclose(par['zH'], [5.0, 6.0, 7.0])
    assert array_len == 3

    with raises(ValueError):
        par, array_keys, array_len = model._parse_parameter_overrides(
            zH=[5.0, 6.0], v=[1.0, 2.0, 3.0])


def test_Meijaard2007Model(show=False):

    parameter_set = Meijaard2007ParameterSet(meijaard2007_parameters, True)

    model = Meijaard2007Model(parameter_set)

    M, C1, K0, K2 = model.form_reduced_canonical_matrices()
    assert M.shape == (2, 2)
    assert C1.shape == (2, 2)
    assert K0.shape == (2, 2)
    assert K2.shape == (2, 2)
    M, C1, K0, K2 = model.form_reduced_canonical_matrices(
        w=np.linspace(0.5, 1.5, num=5))
    assert M.shape == (5, 2, 2)
    assert C1.shape == (5, 2, 2)
    assert K0.shape == (5, 2, 2)
    assert K2.shape == (5, 2, 2)
    M, C1, K0, K2 = model.form_reduced_canonical_matrices(
        w=np.linspace(0.5, 1.5, num=10), v=np.linspace(1, 3, num=10))
    assert M.shape == (10, 2, 2)
    assert C1.shape == (10, 2, 2)
    assert K0.shape == (10, 2, 2)
    assert K2.shape == (10, 2, 2)

    A, B = model.form_state_space_matrices()
    assert A.shape == (4, 4)
    assert B.shape == (4, 2)
    A, B = model.form_state_space_matrices(w=np.linspace(0.5, 1.5, num=5))
    assert A.shape == (5, 4, 4)
    assert B.shape == (5, 4, 2)
    A, B = model.form_state_space_matrices(v=np.linspace(0, 10, num=10))
    assert A.shape == (10, 4, 4)
    assert B.shape == (10, 4, 2)

    A, B = model.form_state_space_matrices(v=np.linspace(0, 10, num=10),
                                           g=np.linspace(0, 10, num=10))
    assert A.shape == (10, 4, 4)
    assert B.shape == (10, 4, 2)

    evals, evecs = model.calc_eigen()
    assert evals.shape == (4,)
    assert evecs.shape == (4, 4)
    evals, evecs = model.calc_eigen(g=6.0)
    assert evals.shape == (4,)
    assert evecs.shape == (4, 4)
    evals, evecs = model.calc_eigen(v=np.linspace(0, 10, num=10))
    assert evals.shape == (10, 4)
    assert evecs.shape == (10, 4, 4)
    model.plot_eigenvalue_parts(v=np.linspace(0, 10, num=10))
    model.plot_eigenvectors(v=np.linspace(0, 10, num=4))
    if show:
        import matplotlib.pyplot as plt
        plt.show()



#!/usr/bin/env python

from math import pi
import numpy as np
from uncertainties import umath, unumpy

# local modules
from .geometry import calculate_l1_l2, fwheel_to_handlebar_ref


def cartesian(arrays, out=None):
    """Generate a cartesian product of input arrays.

    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.

    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.

    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]])

    """

    arrays = [np.asarray(x) for x in arrays]
    dtype = arrays[0].dtype

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = n // arrays[0].size
    out[:, 0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m, 1:])
        for j in range(1, arrays[0].size):
            out[j*m:(j+1)*m, 1:] = out[0:m, 1:]
    return out


def center_of_mass(slopes, intercepts):
    '''Returns the center of mass relative to the slopes and intercepts
    coordinate system.

    Parameters
    ----------
    slopes : ndarray, shape(n,)
        The slope of every line used to calculate the center of mass.
    intercepts : ndarray, shape(n,)
        The intercept of every line used to calculate the center of mass.

    Returns
    -------
    x : float
        The abscissa of the center of mass.
    y : float
        The ordinate of the center of mass.

    '''
    num = range(len(slopes))
    allComb = cartesian((num, num))
    comb = []
    # remove doubles
    for row in allComb:
        if row[0] != row[1]:
            comb.append(row)
    comb = np.array(comb)

    # initialize the matrix to store the line intersections
    lineX = np.zeros((len(comb), 2), dtype='object')
    # for each line intersection...
    for j, row in enumerate(comb):
        sl = np.array([slopes[row[0]], slopes[row[1]]])
        a = unumpy.array(np.vstack((-sl, np.ones((2)))).T)
        b = np.array([intercepts[row[0]], intercepts[row[1]]])
        lineX[j] = np.dot(a.I, b)
    com = np.mean(lineX, axis=0)

    return com[0], com[1]


def com_line(alpha, a, par, part, l1, l2):
    '''Returns the slope and intercept for the line that passes through the
    part's center of mass with reference to the benchmark bicycle coordinate
    system.

    Parameters
    ----------
    alpha : float
        The angle the head tube makes with the horizontal. When looking at the
        bicycle from the right side this is the angle between a vector point
        out upwards along the steer axis and the earth horizontal with the
        positve direction pointing from the left to the right. If the bike is
        in its normal configuration this would be 90 degrees plus the steer
        axis tilt (lambda).
    a : float
        The distance from the pendulum axis to a reference point on the part,
        typically the wheel centers. This is positive if the point falls to the
        left of the axis and negative otherwise.
    par : dictionary
        Benchmark parameters. Must include lam, rR, rF, w
    part : string
        The subscript denoting which part this refers to.
    l1, l2 : floats
        The location of the handlebar reference point relative to the front
        wheel center when the fork is split. This is measured perpendicular to
        and along the steer axis, respectively.

    Returns
    -------
    m : float
        The slope of the line in the benchmark coordinate system.
    b : float
        The z intercept in the benchmark coordinate system.

    '''

    # beta is the angle between the x bike frame and the x pendulum frame,
    # rotation about positive y
    beta = par['lam'] - alpha * pi / 180

    # calculate the slope of the center of mass line
    m = -umath.tan(beta)

    # calculate the z intercept
    # this the bicycle frame
    if part == 'B':
        b = -a / umath.cos(beta) - par['rR']
    # this is the fork (without handlebar) or the fork and handlebar combined
    elif part == 'S' or part == 'H':
        b = -a / umath.cos(beta) - par['rF'] + par['w'] * umath.tan(beta)
    # this is the handlebar (without fork)
    elif part == 'G':
        u1, u2 = fwheel_to_handlebar_ref(par['lam'], l1, l2)
        b = (-a/umath.cos(beta) - (par['rF'] + u2) + (par['w'] -
                                                      u1)*umath.tan(beta))
    else:
        print(part, "doesn't exist")
        raise KeyError

    return m, b, beta


def part_com_lines(mp, par, forkIsSplit):
    '''Returns the slopes and intercepts for all of the center of mass lines
    for each part.

    Parameters
    ----------
    mp : dictionary
        Dictionary with the measured parameters.

    Returns
    -------
    slopes : dictionary
        Contains a list of slopes for each part.
    intercepts : dictionary
        Contains a list of intercepts for each part.

    The slopes and intercepts lists are in order with respect to each other and
    the keyword is either 'B', 'H', 'G' or 'S' for Frame, Handlebar/Fork,
    Handlerbar, and Fork respectively.

    '''
    # find the slope and intercept for pendulum axis
    if forkIsSplit:
        l1, l2 = calculate_l1_l2(mp['h6'], mp['h7'], mp['d5'],
                                 mp['d6'], mp['l'])
        slopes = {'B': [], 'G': [], 'S': []}
        intercepts = {'B': [], 'G': [], 'S': []}
        betas = {'B': [], 'G': [], 'S': []}
    else:
        l1, l2 = 0., 0.
        slopes = {'B': [], 'H': []}
        intercepts = {'B': [], 'H': []}
        betas = {'B': [], 'H': []}

    # get the alpha keys and put them in order
    listOfAlphaKeys = [x for x in mp.keys() if x.startswith('alpha')]
    listOfAlphaKeys.sort()

    # caluclate m, b and beta for each orientation
    for key in listOfAlphaKeys:
        alpha = mp[key]
        a = mp['a' + key[5:]]
        part = key[5]
        m, b, beta = com_line(alpha, a, par, part, l1, l2)
        slopes[part].append(m)
        intercepts[part].append(b)
        betas[part].append(beta)

    return slopes, intercepts, betas


def total_com(coordinates, masses):
    '''Returns the center of mass of a group of objects if the indivdual
    centers of mass and mass is provided.

    coordinates : array_like, shape(3,n)
        The rows are the x, y and z coordinates, respectively and the columns
        are for each object.
    masses : array_like, shape(3,)
        An array of the masses of multiple objects, the order should correspond
        to the columns of coordinates.

    Returns
    -------
    mT : float
        Total mass of the objects.
    cT : ndarray, shape(3,)
        The x, y, and z coordinates of the total center of mass.

    '''
    products = np.asarray(masses) * np.asarray(coordinates)
    mT = np.sum(masses)
    cT = np.sum(products, axis=1) / mT
    return mT, cT



from abc import ABC
import warnings

import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate as spi

from .bicycle import benchmark_par_to_canonical, ab_matrix, sort_eigenmodes


class _Model(ABC):
    """A model is a set of differential algebraic equations in time that have:
    constants and time varying (coordinates, speeds, and exogenous inputs).
    The model can be nonlinear, linear, have algebraic constraints, or not.  A
    parameter set is associated with a particular parameterization of one or
    more models.

    """
    pass


class Meijaard2007Model(_Model):
    """Whipple-Carvallo model presented in [Meijaard2007]_. It is both linear
    and the minimal model in terms of states and coordinates that fully
    describe the vehicle's dynamics: self-stability and non-minimum phase
    behavior.

    Parameters
    ==========
    parameter_set : ParameterSet
        The ``paramter_set.to_parameterization('meijaard2007')`` must
        return a dictionary that maps floats to the parameter keys
        containing:

        - ``IBxx`` : x moment of inertia of the frame/rider [kg*m**2]
        - ``IBxz`` : xz product of inertia of the frame/rider [kg*m**2]
        - ``IBzz`` : z moment of inertia of the frame/rider [kg*m**2]
        - ``IFxx`` : x moment of inertia of the front wheel [kg*m**2]
        - ``IFyy`` : y moment of inertia of the front wheel [kg*m**2]
        - ``IHxx`` : x moment of inertia of the handlebar/fork [kg*m**2]
        - ``IHxz`` : xz product of inertia of the handlebar/fork [kg*m**2]
        - ``IHzz`` : z moment of inertia of the handlebar/fork [kg*m**2]
        - ``IRxx`` : x moment of inertia of the rear wheel [kg*m**2]
        - ``IRyy`` : y moment of inertia of the rear wheel [kg*m**2]
        - ``c`` : trail [m]
        - ``g`` : acceleration due to gravity [m/s**2]
        - ``lam`` : steer axis tilt [rad]
        - ``mB`` : frame/rider mass [kg]
        - ``mF`` : front wheel mass [kg]
        - ``mH`` : handlebar/fork assembly mass [kg]
        - ``mR`` : rear wheel mass [kg]
        - ``rF`` : front wheel radius [m]
        - ``rR`` : rear wheel radius [m]
        - ``v`` : speed [m/s]
        - ``w`` : wheelbase [m]
        - ``xB`` : x distance to the frame/rider center of mass [m]
        - ``xH`` : x distance to the frame/rider center of mass [m]
        - ``zB`` : z distance to the frame/rider center of mass [m]
        - ``zH`` : z distance to the frame/rider center of mass [m]

    Attributes
    ==========
    input_vars : list of strings
        Ordered list of ASCII strings that name the model's input variables.
    state_vars : list of strings
        Ordered list of ASCII strings that name the model's state variables.
    input_vars_latex : list of raw strings
        Ordered list of LaTeX strings that name the model's input variables.
    state_vars_latex : list of raw strings
        Ordered list of LaTeX strings that name the model's state variables.

    References
    ==========

    .. [Meijaard2007] Meijaard J.P, Papadopoulos Jim M, Ruina Andy and Schwab
       A.L, 2007, Linearized dynamics equations for the balance and steer of a
       bicycle: a benchmark and review, Proc. R. Soc. A., 463:1955–1982
       http://doi.org/10.1098/rspa.2007.1857

    """
    input_vars = ['Tphi', 'Tdelta']
    state_vars = ['phi', 'delta', 'phidot', 'deltadot']
    input_vars_latex = [r'T_\phi', r'T_\delta']
    state_vars_latex = [r'\phi', r'\delta', r'\dot{\phi}', r'\dot{\delta}']

    def __init__(self, parameter_set):
        self.parameter_set = parameter_set.to_parameterization('Meijaard2007')

    def _parse_parameter_overrides(self, **parameter_overrides):
        """Returns the model's parameter dictionary with the overridden
        parameters replaced.

        Parameters
        ==========
        parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Returns
        =======
        par : dictionary
            Copy of self.parameter_set.parameters with overridden parameter
            values.
        array_keys : list
            All parameter key strings that hold arrays.
        array_len : None or int
            If there are arrays, this is the common length.

        """

        par = self.parameter_set.parameters.copy()

        array_len = None
        array_keys = []

        for key, val in parameter_overrides.items():
            if key not in par.keys():  # don't add invalid keys
                msg = '{} is not a valid parameter, ignoring'
                warnings.warn(msg.format(key))
            else:
                if np.isscalar(val):
                    par[key] = float(val)
                else:  # is an array
                    if array_len is None:
                        array_len = len(val)
                    if len(val) != array_len:
                        msg = ('All array valued parameters must have the '
                               'same length.')
                        raise ValueError(msg)
                    array_keys.append(key)
                    par[key] = val

        return par, array_keys, array_len

    def form_reduced_canonical_matrices(self, **parameter_overrides):
        """Returns the canonical speed and gravity independent matrices for the
        Whipple-Carvallo bicycle model linearized about the nominal upright
        configuration.

        Parameters
        ==========
        **parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Returns
        =======
        M : ndarray, shape(2,2) or shape(n,2,2)
            Mass matrix.
        C1 : ndarray, shape(2,2) or shape(n,2,2)
            Velocity independent damping matrix.
        K0 : ndarray, shape(2,2) or shape(n,2,2)
            Gravity independent part of the stiffness matrix.
        K2 : ndarray, shape(2,2) or shape(n,2,2)
            Velocity squared independent part of the stiffness matrix.

        Notes
        =====
        The canonical matrices complete the following equation:

        ``M*q'' + v*C1*q' + [g*K0 + v**2*K2]*q = f``

        where:

        - ``q = [phi, delta]``
        - ``f = [Tphi, Tdelta]``

        ``phi``
            Bicycle roll angle.
        ``delta``
            Steer angle.
        ``Tphi``
            Roll torque.
        ``Tdelta``
            Steer torque.
        ``v``
            Bicylce longitudinal speed.
        ``g``
            Acceleration due to gravity.

        Examples
        ========

        >>> from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
        >>> from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
        >>> from bicycleparameters.models import Meijaard2007Model
        >>> p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
        >>> m = Meijaard2007Model(p)
        >>> M, C1, K0, K2 = m.form_reduced_canonical_matrices()
        >>> M
        array([[102.78013216,   1.53582801],
               [  1.53582801,   0.24890226]])
        >>> C1
        array([[ 0.       , 26.3947333],
               [-0.4503006,  1.037066 ]])
        >>> K0
        array([[-89.32195981,  -1.74159477],
               [ -1.74159477,  -0.67769624]])
        >>> K2
        array([[ 0.        , 74.12543   ],
               [ 0.        ,  1.57021553]])
        >>> M, _, _, _ = m.form_reduced_canonical_matrices(mB=150.0)
        >>> M
        array([[176.52178763,   2.69074048],
               [  2.69074048,   0.26699004]])

        """
        par, array_keys, array_len = self._parse_parameter_overrides(
            **parameter_overrides)

        mutable_par = par.copy()

        # these are the only variables needed to calculate M, C1, K0, K2
        canon_deps = ['IBxx', 'IBxz', 'IByy', 'IBzz',
                      'IFxx', 'IFyy',
                      'IHxx', 'IHxz', 'IHyy', 'IHzz',
                      'IRxx', 'IRyy',
                      'c', 'lam',
                      'mB', 'mF', 'mH', 'mR',
                      'rF', 'rR', 'w',
                      'xB', 'xH', 'zB', 'zH']

        compute_arrays = False
        for key in array_keys:
            if key in canon_deps:
                compute_arrays = True
                break

        if array_keys and compute_arrays:
            n = array_len
            M = np.zeros((n, 2, 2))
            C1 = np.zeros((n, 2, 2))
            K0 = np.zeros((n, 2, 2))
            K2 = np.zeros((n, 2, 2))
            for i in range(n):
                for key in array_keys:
                    mutable_par[key] = par[key][i]
                M[i], C1[i], K0[i], K2[i] = benchmark_par_to_canonical(
                    mutable_par)
            return M, C1, K0, K2
        else:
            return benchmark_par_to_canonical(par)

    def form_state_space_matrices(self, **parameter_overrides):
        """Returns the A and B matrices for the Whipple-Carvallo model
        linearized about the upright constant velocity configuration.

        Parameters
        ==========
        **parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Returns
        =======
        A : ndarray, shape(4,4) or shape(n,4,4)
            The state matrix.
        B : ndarray, shape(4,2) or shape(n,4,2)
            The input matrix.

        Notes
        =====
        ``A`` and ``B`` describe the Whipple model in state space form:

        ``x' = A * x + B * u``

        where the states are::

            x = |roll angle | = |phi     |
                |steer angle|   |delta   |
                |roll rate  |   |phidot  |
                |steer rate |   |deltadot|

        and the inputs are::

            u = |roll torque | = |Tphi  |
                |steer torque|   |Tdelta|

        Examples
        ========

        >>> from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
        >>> from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
        >>> from bicycleparameters.models import Meijaard2007Model
        >>> p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
        >>> m = Meijaard2007Model(p)
        >>> A, B = m.form_state_space_matrices()
        >>> A
        array([[ 0.        ,  0.        ,  1.        ,  0.        ],
               [ 0.        ,  0.        ,  0.        ,  1.        ],
               [ 8.26150335, -0.9471634 , -0.02977958, -0.21430735],
               [17.66475151, 26.24590352,  1.99289841, -2.84419587]])
        >>> B
        array([[ 0.        ,  0.        ],
               [ 0.        ,  0.        ],
               [ 0.01071772, -0.06613267],
               [-0.06613267,  4.42570676]])

        """
        par, array_keys, array_len = self._parse_parameter_overrides(
            **parameter_overrides)

        M, C1, K0, K2 = self.form_reduced_canonical_matrices(
            **parameter_overrides)

        if len(M.shape) == 3:  # one of the parameters (not v or g) is an array
            A = np.zeros((M.shape[0], 4, 4))
            B = np.zeros((M.shape[0], 4, 2))
            for i, (Mi, C1i, K0i, K2i) in enumerate(zip(M, C1, K0, K2)):
                if 'g' in array_keys:
                    g = par['g'][i]
                else:
                    g = par['g']
                if 'v' in array_keys:
                    v = par['v'][i]
                else:
                    v = par['v']
                A[i], B[i] = ab_matrix(Mi, C1i, K0i, K2i, v, g)
        elif 'v' in array_keys or 'g' in array_keys:
            n = array_len
            A = np.zeros((n, 4, 4))
            B = np.zeros((n, 4, 2))
            for i in range(n):
                if 'g' in array_keys:
                    g = par['g'][i]
                else:
                    g = par['g']
                if 'v' in array_keys:
                    v = par['v'][i]
                else:
                    v = par['v']
                A[i], B[i] = ab_matrix(M, C1, K0, K2, v, g)
        else:  # scalar parameters
            A, B = ab_matrix(M, C1, K0, K2, par['v'], par['g'])

        return A, B

    def calc_eigen(self, left=False, **parameter_overrides):
        """Returns the right (or left) eigenvalues and eigenvectors of the
        linear model.

        Parameters
        ==========
        left : boolean, optional
            If true, the left eigenvectors will be returned, i.e.
            ``A.T*v=lam*v``.
        **parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Returns
        =======
        evals : ndarray, shape(4,) or shape (n,4)
            Eigenvalues.
        evecs : ndarray, shape(4,4) or shape (n,4,4)
            Eigenvectors, each columns are eigenvectors and are associated with
            same index of the eigenvalues.

        Examples
        ========

        >>> from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
        >>> from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
        >>> from bicycleparameters.models import Meijaard2007Model
        >>> p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
        >>> m = Meijaard2007Model(p)
        >>> evals, evecs = m.calc_eigen()
        >>> evals
        array([-6.74423162+0.j        , -2.9146438 +0.j        ,
                3.39244999+0.61085077j,  3.39244999-0.61085077j])
        >>> evecs
        array([[ 0.00197344+0.j        , -0.2953538 +0.j        ,
                 0.04320146-0.0753826j ,  0.04320146+0.0753826j ],
               [ 0.14665803+0.j        ,  0.13447333+0.j        ,
                -0.26053575+0.04691255j, -0.26053575-0.04691255j],
               [-0.01330934+0.j        ,  0.86085111+0.j        ,
                 0.1926063 -0.22934205j,  0.1926063 +0.22934205j],
               [-0.98909574+0.j        , -0.39194186+0.j        ,
                -0.91251108+0.j        , -0.91251108-0.j        ]])

        """
        A, B = self.form_state_space_matrices(**parameter_overrides)

        if len(A.shape) == 3:  # array version
            evals = np.zeros(A.shape[:2], dtype='complex128')
            evecs = np.zeros(A.shape, dtype='complex128')
            for i, Ai in enumerate(A):
                if left:
                    Ai = np.transpose(Ai)
                evals[i], evecs[i] = np.linalg.eig(Ai)
            return evals, evecs
        else:
            if left:
                A = np.transpose(A)
            return np.linalg.eig(A)

    def _calc_modal_controllability(self, acute=True, **parameter_overrides):
        """Returns the modal controllability [1]_ measures for each input and
        each eigenmode. The modal controllability is defined as the angle
        between each left eigenvector and each input column.

        Parameters
        ==========
        acute : boolean, optional
           If true only angles from 0 to pi/2 will be returned from the
           arrcos() computation. If false, angles from -pi/2 to pi/2 will be
           returned.
        **parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Returns
        =======
        beta : ndarray, shape(4,2) or shape(n, 4,2)
            Modal controllability angle for each eigenmode and each input.

        Notes
        =====

        ``x' = A*x + B*u``

        - A is n x n
        - B is n x m

        The columns of B are associated with the jth input:

        ``B = [b1, ..., bj, ..., bm]``

        The columns of Q are ith left eigenvectors of A, i.e. _, ``Q =
        eig(A.T)``:

        ``Q = [q1, ..., qi, ..., qn]``

        The modal controllability angle beta_ij is defined as::

           cos(beta_ij) = |qi.T @ bj|
                          -------------
                          ||qi||*||bj||

        - qi : ith left eigenvector
        - bj : jth column of B

        When qi is complex, the angle calculation is [2]_::

           cos(beta_ij) = Re(qi.H @ bj)
                          -------------
                          ||qi||*||bj||

        References
        ==========

        .. [1] A. M. A. Hamdan and A. H. Nayfeh, "Measures of modal
           controllability and observability for first- and second-order linear
           systems," Journal of Guidance, Control, and Dynamics, vol. 12, no.
           3, pp. 421–428, 1989, doi: 10.2514/3.20424.

        .. [2] https://en.wikipedia.org/wiki/Dot_product#Complex_vectors


        """

        _, B = self.form_state_space_matrices(**parameter_overrides)
        evals, evecs = self.calc_eigen(left=True, **parameter_overrides)
        evals, evecs = sort_eigenmodes(evals, evecs)

        def mod_cont(q, b, acute=True):
            """Returns the modal controllability value of the eigenvector q and
            input matrix column b.

            Parameters
            ==========
            q : array_like, shape(n,)
                A complex left eigenvector.
            b : array_like, shape(n,)
                A real column of the input matrix.

            Returns
            =======
            beta : float
                The (acute) angle in radians between q and b.

            """
            # vdot takes the complex conjugate of the first argument before
            # taking the dot product.
            num = np.real(np.vdot(q, b))  # Re(q.H @ b)
            # num = np.real(b @ np.conjugate(q).T)

            # norm_q = np.abs(np.sqrt(np.conjugate(q).T @ q))
            # norm_b = np.sqrt(b.T @ b)
            # den = norm_q*norm_b

            # norm() returns a real valued answer for the 2-norm
            den = np.linalg.norm(q)*np.linalg.norm(b)

            # NOTE : abs() forces 0 to 90 deg instead of 0 to 180 deg, i.e.
            # always the acute angle.
            if acute:
                cosbeta = np.abs(num/den)
            else:
                cosbeta = num/den
            return np.arccos(cosbeta)

        if len(B.shape) == 3:  # array version
            mod_ctrb = np.empty_like(B)
            for k, (Bk, vk) in enumerate(zip(B, evecs)):
                # columns of the evecs and columns of B
                for i, vi in enumerate(vk.T):
                    for j, bj in enumerate(Bk.T):
                        mod_ctrb[k, i, j] = mod_cont(vi, bj, acute=acute)
        else:
            mod_ctrb = np.empty((evecs.shape[1], B.shape[1]))
            for i, vi in enumerate(evecs.T):
                for j, bj in enumerate(B.T):
                    mod_ctrb[i, j] = mod_cont(vi, bj, acute=acute)

        return mod_ctrb

    def _plot_modal_controllability(self, axes=None, acute=True,
                                    **parameter_overrides):
        """Returns axes shape(4,2) with plots of the modal controllability for
        each input and each eigenmode."""

        par, array_keys, array_len = self._parse_parameter_overrides(
            **parameter_overrides)

        betas = self._calc_modal_controllability(acute=acute,
                                                 **parameter_overrides)
        betas = np.rad2deg(betas)

        if axes is None:
            fig, axes = plt.subplots(*betas[0].shape, sharex=True)

        for i, row in enumerate(axes):
            row[0].set_ylabel('Mode {}'.format(i + 1))
            for j, col in enumerate(row):
                col.plot(par[array_keys[0]], betas[:, i, j])

        axes[3, 0].set_xlabel(array_keys[0])
        axes[3, 1].set_xlabel(array_keys[0])

        axes[0, 0].set_title(r'Input: $T_\phi$')
        axes[0, 1].set_title(r'Input: $T_\delta$')

        return axes

    def plot_eigenvalue_parts(self, ax=None, colors=None,
                              **parameter_overrides):
        """Returns a matplotlib axis of the real and imaginary parts of the
        eigenvalues plotted against the provided parameter.

        Parameters
        ==========
        ax : Axes
            Matplotlib axes.
        colors : sequence, len(4)
            Matplotlib colors for the 4 modes.
        **parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           import numpy as np
           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           from bicycleparameters.models import Meijaard2007Model
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           m = Meijaard2007Model(p)
           m.plot_eigenvalue_parts(v=np.linspace(0.0, 10.0, num=101))

        """

        if ax is None:
            fig, ax = plt.subplots()

        evals, evecs = self.calc_eigen(**parameter_overrides)
        if len(evals.shape) > 1:
            evals, evecs = sort_eigenmodes(evals, evecs)
        else:
            evals, evecs = [evals], [evecs]

        par, array_keys, _ = self._parse_parameter_overrides(
            **parameter_overrides)

        if colors is None:
            colors = ['C0', 'C1', 'C2', 'C3']
        legend = ['Mode 1', 'Mode 2', 'Mode 3', 'Mode 4',
                  'Mode 1', 'Mode 2', 'Mode 3', 'Mode 4']
        # imaginary components
        for eval_sequence, color, label in zip(evals.T, colors, legend):
            ax.plot(par[array_keys[0]], np.abs(np.imag(eval_sequence)),
                    color=color, label=label, linestyle='--')

        # plot the real parts of the eigenvalues
        for eval_sequence, color, label in zip(evals.T, colors, legend):
            ax.plot(par[array_keys[0]], np.real(eval_sequence), color=color,
                    label=label)

        # set labels and limits
        ax.set_ylabel('Real and Imaginary Parts of the Eigenvalue [1/s]')
        ax.set_xlim((par[array_keys[0]][0], par[array_keys[0]][-1]))

        ax.grid()

        ax.set_xlabel(array_keys[0])

        return ax

    def plot_eigenvectors(self, **parameter_overrides):
        """Plots the components of the eigenvectors in the real and imaginary
        plane.

        Parameters
        ==========
        **parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Returns
        =======
        axes : ndarray, shape(n, 4)
            Polar plot axes for each eigenvector (columns). The rows correspond
            to a varied parameter.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           import numpy as np
           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           from bicycleparameters.models import Meijaard2007Model
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           m = Meijaard2007Model(p)
           m.plot_eigenvectors(v=[1.0, 3.0, 5.0])

        """
        par, arr_keys, _ = self._parse_parameter_overrides(
            **parameter_overrides)
        states = [r'\phi', r'\delta', r'\dot{\phi}', r'\dot{\delta}']

        eval_seq, evec_seq = self.calc_eigen(**parameter_overrides)
        eval_seq, evec_seq = np.atleast_2d(eval_seq), np.atleast_3d(evec_seq)

        if eval_seq.shape[0] > 10:
            msg = ('Plots will be too large, use fewer than 11 values in the '
                   'varied parameter.')
            raise ValueError(msg)

        # TODO : sort_eigenmodes() is doing something funny and not adding the
        # 4th eigenvalue, so you often end up with duplicates eigenvalues for
        # one eigenval and one missing. Also the algorithm may not work well
        # with spaced out eigenvalues, which is what I've been trying here. You
        # may have to calculate eigenvals/vec across closer spacing, then
        # sample out the ones you want. For now, we don't sort coarse spaced
        # eigenvalues.
        # if arr_keys:
            # eval_seq, evec_seq = sort_eigenmodes(eval_seq, evec_seq)

        fig, axes = plt.subplots(*eval_seq.shape,
                                 subplot_kw={'projection': 'polar'})
        axes = np.atleast_2d(axes)
        fig.set_size_inches(axes.shape[1]*3, axes.shape[0]*3)
        lw = list(range(1, len(states) + 1))
        lw.reverse()

        for k, (evals, evecs, par_val) in enumerate(zip(eval_seq, evec_seq,
                                                        par[arr_keys[0]])):

            axes[k, 0].set_ylabel('{} = {:1.2f}'.format(arr_keys[0], par_val),
                                  labelpad=30)

            for i, (eigenval, eigenvec) in enumerate(zip(evals, evecs.T)):

                max_com = np.abs(eigenvec[:2]).max()

                for j, component in enumerate(eigenvec[:2]):

                    radius = np.abs(component)/max_com
                    theta = np.angle(component)
                    axes[k, i].plot([0, theta], [0, radius], lw=lw[j])

                axes[k, i].set_rmax(1.0)
                msg = r'Eigenvalue: {:1.3f}'
                if eigenval.real >= 0.0:
                    fontcolor = 'red'  # red indicates unstable
                else:
                    fontcolor = 'black'
                axes[k, i].set_title(msg.format(eigenval),
                                     fontdict={'color': fontcolor})

        axes[0, 0].legend(['$' + s + '$' for s in states],
                          loc='upper center', bbox_to_anchor=(0.5, 1.05),
                          fancybox=True, shadow=True, ncol=4)

        fig.tight_layout()

        return axes

    def simulate(self, times, initial_conditions, input_func=None,
                 **parameter_overrides):
        """Returns the state and input trajectories at each time value.

        Parameters
        ==========
        times : array_like, shape(n,)
            Monotonic increasing time values to simulate over.
        initial_conditions : array_like, shape(4,)
            Initial values of the states.
        input_func : function
            Takes form u = f(t, x) where u is array_like, shape(2,).
        **parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Returns
        =======
        states : ndarray, shape(n, 4)
            State trajectories over n time values.
        inputs : ndatrray, shape(n, 2)
            Input trajectories over n time values.

        """

        par, arr_keys, _ = self._parse_parameter_overrides(
            **parameter_overrides)

        if arr_keys:
            raise ValueError('Can only simulate with fixed parameters.')

        A, B = self.form_state_space_matrices(**parameter_overrides)

        if input_func is None:
            def eval_rhs(t, x):
                return A@x
        else:
            def eval_rhs(t, x):
                return A@x + B@input_func(t, x)

        res = spi.solve_ivp(eval_rhs,
                            (times[0], times[-1]),
                            initial_conditions,
                            t_eval=times)

        if input_func is None:
            inputs = np.zeros((len(times), 2))
        else:
            inputs = np.empty((len(times), 2))
            for i, ti in enumerate(times):
                ui = input_func(ti, res.y[:, i])
                inputs[i, :] = ui[:]

        return res.y.T, inputs

    def plot_simulation(self, times, initial_conditions, input_func=None,
                        **parameter_overrides):
        """Returns the state and input trajectories at each time value.

        Parameters
        ==========
        times : array_like, shape(n,)
            Monotonic increasing time values to simulate over.
        initial_conditions : array_like, shape(4,)
            Initial values of the states.
        input_func : function
            Takes form u = f(t, x) where u is array_like, shape(2,).
        **parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Returns
        =======
        axes : ndarray, shape(3,)
            Three subplots that plot the input trajectories, state angle
            trajectories, and state angular rates.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           import numpy as np
           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           from bicycleparameters.models import Meijaard2007Model
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           m = Meijaard2007Model(p)
           times = np.linspace(0.0, 5.0, num=51)
           x0 = np.deg2rad([10.0, 5.0, 0.0, 0.0])
           m.plot_simulation(times, x0, v=6.0)

        """
        res, inputs = self.simulate(times, initial_conditions,
                                    input_func=input_func,
                                    **parameter_overrides)

        fig, axes = plt.subplots(3, sharex=True)

        axes[0].plot(times, inputs)
        axes[0].legend([r'$T_\phi$', r'$T_\delta$'])
        axes[1].plot(times, np.rad2deg(res[:, :2]))
        axes[1].legend(['$' + lab + '$' for lab in self.state_vars_latex[:2]])
        axes[2].plot(times, np.rad2deg(res[:, 2:]))
        axes[2].legend(['$' + lab + '$' for lab in self.state_vars_latex[2:]])

        axes[2].set_xlabel('Time [s]')

        return axes

    def simulate_modes(self, times, **parameter_overrides):
        """Returns simulation results showing the behavior of each
        eigenmode.

        Parameters
        ==========
        times : array_like, shape(n,)
            Monotonic increasing time values to simulate over.
        **parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Returns
        =======
        results : ndarray, shape(4, n, 4)
            State trajectories for each mode with the shape corresponding to
            (mode, time, state).

        """

        par, arr_keys, _ = self._parse_parameter_overrides(
            **parameter_overrides)

        if arr_keys:
            raise ValueError('Can only simulate with fixed parameters.')

        A, B = self.form_state_space_matrices(**parameter_overrides)
        evals, evecs = self.calc_eigen(**parameter_overrides)

        def eval_rhs(t, x):
            return A@x

        results = np.empty((4, len(times), 4))

        for i, evec in enumerate(evecs.T):
            initial_condition = evec.real

            sim_res = spi.solve_ivp(eval_rhs, (times[0], times[-1]),
                                    initial_condition, t_eval=times)
            results[i] = sim_res.y.T

        return results

    def plot_mode_simulations(self, times, **parameter_overrides):
        """Returns matplotlib subplot axes with a simulation of each mode.

        Parameters
        ==========
        times : array_like, shape(n,)
            Monotonic increasing time values to simulate over.
        **parameter_overrides : dictionary
            Parameter keys that map to floats or array_like of floats
            shape(n,). All keys that map to array_like must be of the same
            length.

        Returns
        =======
        axes : ndarray, shape(4,2)
            Subplot axes with the modes on the rows and the angles in the first
            column and the angular rates in the second column.

        Examples
        ========

        .. plot::
           :include-source: True
           :context: reset

           import numpy as np
           from bicycleparameters.parameter_dicts import meijaard2007_browser_jason
           from bicycleparameters.parameter_sets import Meijaard2007ParameterSet
           from bicycleparameters.models import Meijaard2007Model
           p = Meijaard2007ParameterSet(meijaard2007_browser_jason, True)
           m = Meijaard2007Model(p)
           times = np.linspace(0.0, 5.0, num=51)
           m.plot_mode_simulations(times, v=6.0)

        """
        results = self.simulate_modes(times, **parameter_overrides)

        fig, axes = plt.subplots(4, 2, sharex=True)

        for i, res in enumerate(results):
            axes[i, 0].plot(times, np.rad2deg(results[i, :, :2]))
            axes[i, 0].legend(['$' + lab + '$'
                               for lab in self.state_vars_latex[:2]])
            axes[i, 1].plot(times, np.rad2deg(results[i, :, 2:]))
            axes[i, 1].legend(['$' + lab + '$'
                               for lab in self.state_vars_latex[2:]])

        axes[3, 0].set_xlabel('Time [s]')
        axes[3, 1].set_xlabel('Time [s]')

        return axes



#!/usr/bin/env python

from math import pi
from uncertainties import unumpy, umath, UFloat
import numpy as np

# local modules
from .com import total_com


def combine_bike_rider(bicyclePar, riderPar):
    """
    Combines the inertia of the bicycle frame with the
    inertia of a rider.

    Parameters
    ----------
    bicyclePar : dictionary
        The benchmark parameter set of a bicycle.
    riderPar : dictionary
        The rider's mass, center of mass, and inertia expressed in the
        benchmark bicycle reference frame.

    Returns
    -------
    bicyclePar : dictionary
        The benchmark bicycle parameters with a rigid rider added to the
        bicycle frame.

    """

    # list the masses of the rider and bicycle
    masses = np.array([riderPar['mB'], bicyclePar['mB']])
    # list the centers of mass of the rider and bicycle
    coordinates = np.array([[riderPar['xB'], bicyclePar['xB']],
                            [riderPar['yB'], 0.],
                            [riderPar['zB'], bicyclePar['zB']]])
    # calculate the new mass and center of mass
    mT, cT = total_com(coordinates, masses)
    # get inertia tensors for the bicycle and rider
    IRider = part_inertia_tensor(riderPar, 'B')
    IBicycle = part_inertia_tensor(bicyclePar, 'B')
    # calculate the distance from the center of mass of each body to the
    # center of mass of the combined body
    dRider = np.array([riderPar['xB'] - cT[0],
                       riderPar['yB'] - cT[1],
                       riderPar['zB'] - cT[2]])
    dBicycle = np.array([bicyclePar['xB'] - cT[0],
                         0.,
                         bicyclePar['zB'] - cT[2]])
    # calculate the total inertia about the total body center of mass
    I = (parallel_axis(IRider, riderPar['mB'], dRider) +
         parallel_axis(IBicycle, bicyclePar['mB'], dBicycle))
    # assign new inertia back to bike
    bicyclePar['xB'] = cT[0]
    bicyclePar['zB'] = cT[2]
    bicyclePar['yB'] = 0.0
    bicyclePar['mB'] = mT
    bicyclePar['IBxx'] = I[0, 0]
    bicyclePar['IBxz'] = I[0, 2]
    bicyclePar['IByy'] = I[1, 1]
    bicyclePar['IBzz'] = I[2, 2]

    return bicyclePar


def compound_pendulum_inertia(m, g, l, T):
    '''Returns the moment of inertia for an object hung as a compound
    pendulum.

    Parameters
    ----------
    m : float
        Mass of the pendulum.
    g : float
        Acceration due to gravity.
    l : float
        Length of the pendulum.
    T : float
        The period of oscillation.

    Returns
    -------
    I : float
        Moment of interia of the pendulum.

    '''

    I = (T / 2. / pi)**2. * m * g * l - m * l**2.

    return I


def inertia_components(jay, beta):
    '''Returns the 2D orthogonal inertia tensor.

    When at least three moments of inertia and their axes orientations are
    known relative to a common inertial frame of a planar object, the orthoganl
    moments of inertia relative the frame are computed.

    Parameters
    ----------
    jay : ndarray, shape(n,)
        An array of at least three moments of inertia. (n >= 3)
    beta : ndarray, shape(n,)
        An array of orientation angles corresponding to the moments of inertia
        in jay.

    Returns
    -------
    eye : ndarray, shape(3,)
        Ixx, Ixz, Izz

    '''
    sb = unumpy.sin(beta)
    cb = unumpy.cos(beta)
    betaMat = unumpy.array(np.vstack((cb**2, -2 * sb * cb, sb**2)).T)
    eye = np.squeeze(np.asarray(np.dot(betaMat.I, jay)))
    return eye


def parallel_axis(Ic, m, d):
    '''Returns the moment of inertia of a body about a different point.

    Parameters
    ----------
    Ic : ndarray, shape(3,3)
        The moment of inertia about the center of mass of the body with respect
        to an orthogonal coordinate system.
    m : float
        The mass of the body.
    d : ndarray, shape(3,)
        The distances along the three ordinates that located the new point
        relative to the center of mass of the body.

    Returns
    -------
    I : ndarray, shape(3,3)
        The moment of inertia about of the body about a point located by d.

    '''
    a = d[0]
    b = d[1]
    c = d[2]
    dMat = np.zeros((3, 3), dtype=object)
    dMat[0] = np.array([b**2 + c**2, -a * b, -a * c])
    dMat[1] = np.array([-a * b, c**2 + a**2, -b * c])
    dMat[2] = np.array([-a * c, -b * c, a**2 + b**2])
    return Ic + m * dMat


def part_inertia_tensor(par, part):
    '''Returns an inertia tensor for a particular part for the benchmark
    parameter set.

    Parameters
    ----------
    par : dictionary
        Complete Benchmark parameter set.
    part : string
        Either 'B', 'H', 'F', 'R', 'G', 'S', 'D'

    Returns
    -------
    I : ndarray, shape(3,3)
        Inertia tensor for the part.

    Notes
    -----
    Parts G, S, and D are additional parts not included in the published paper
    on the benchmark bicycle, they are only relavant if used. See the
    documentation.

    '''
    if isinstance(par['mB'], UFloat):
        dtype = object
    else:
        dtype = 'float64'
    I = np.zeros((3, 3), dtype=dtype)
    # front or rear wheel
    if part in 'FRD':
        axes = np.array([['xx', None, None],
                         [None, 'yy', None],
                         [None, None, 'xx']])
    # all other parts
    else:
        axes = np.array([['xx', None, 'xz'],
                         [None, 'yy', None],
                         ['xz', None, 'zz']])
    for i, row in enumerate(axes):
        for j, col in enumerate(row):
            if col is not None:
                I[i, j] = par['I' + part + col]
    return I


def principal_axes(I):
    '''Returns the principal moments of inertia and the orientation.

    Parameters
    ----------
    I : ndarray, shape(3,3)
        An inertia tensor.

    Returns
    -------
    Ip : ndarray, shape(3,)
        The principal moments of inertia. This is sorted smallest to largest.
    C : ndarray, shape(3,3)
        The rotation matrix.

    '''
    Ip, C = np.linalg.eig(I)
    indices = np.argsort(Ip)
    Ip = Ip[indices]
    C = C.T[indices]
    return Ip, C


def rotate_inertia_tensor(I, angle):
    '''Returns inertia tensor rotated through angle. Only for 2D'''
    ca = umath.cos(angle)
    sa = umath.sin(angle)
    C = np.array([[ca, 0., -sa],
                  [0., 1., 0.],
                  [sa, 0., ca]])
    Irot = np.dot(C, np.dot(I, C.T))
    return Irot


def tor_inertia(k, T):
    '''Calculate the moment of inertia for an ideal torsional pendulm

    Parameters
    ----------
    k: torsional stiffness
    T: period

    Returns
    -------
    I: moment of inertia

    '''

    I = k * T**2 / 4. / pi**2

    return I


def torsional_pendulum_stiffness(I, T):
    '''Calculate the stiffness of a torsional pendulum with a known moment of
    inertia.

    Parameters
    ----------
    I : moment of inertia
    T : period

    Returns
    -------
    k : stiffness

    '''
    k = 4. * I * pi**2 / T**2
    return k


def tube_inertia(l, m, ro, ri):
    '''Calculate the moment of inertia for a tube (or rod) where the x axis is
    aligned with the tube's axis.

    Parameters
    ----------
    l : float
        The length of the tube.
    m : float
        The mass of the tube.
    ro : float
        The outer radius of the tube.
    ri : float
        The inner radius of the tube. Set this to zero if it is a rod instead
        of a tube.

    Returns
    -------
    Ix : float
        Moment of inertia about tube axis.
    Iy, Iz : float
        Moment of inertia about normal axis.

    '''
    Ix = m / 2. * (ro**2 + ri**2)
    Iy = m / 12. * (3 * ro**2 + 3 * ri**2 + l**2)
    Iz = Iy
    return Ix, Iy, Iz



#!/usr/bin/env/ python

import os
from math import pi

import numpy as np
from numpy import ma
from scipy.optimize import leastsq
import matplotlib.pyplot as plt
from uncertainties import ufloat

# local modules
from .io import load_pendulum_mat_file


def average_rectified_sections(data):
    '''Returns a slice of an oscillating data vector based on the max and min
    of the mean of the sections created by retifiying the data.

    Parameters
    ----------
    data : ndarray, shape(n,)

    Returns
    -------
    data : ndarray, shape(m,)
        A slice where m is typically less than n.

    Notes
    -----
    This is a function to try to handle the fact that some of the data from the
    torsional pendulum had a beating like phenomena and we only want to select
    a section of the data that doesn't seem to exhibit the phenomena.

    '''
    # subtract the mean so that there are zero crossings
    meanSubData = data - np.mean(data)
    # find the zero crossings
    zeroCrossings = np.where(np.diff(np.sign(meanSubData)))[0]
    # add a zero to the beginning
    crossings = np.concatenate((np.array([0]), zeroCrossings))
    # find the mean value of the rectified sections and the local indice
    secMean = []
    localMeanInd = []
    for sec in np.split(np.abs(meanSubData), zeroCrossings):
        localMeanInd.append(np.argmax(sec))
        secMean.append(np.mean(sec))
    meanInd = []
    # make the global indices
    for i, val in enumerate(crossings):
        meanInd.append(val + localMeanInd[i])
    # only take the top part of the data because some the zero crossings can be
    # a lot at one point mainly due to the resolution of the daq box
    threshold = np.mean(secMean)
    secMeanOverThresh = []
    indice = []
    for i, val in enumerate(secMean):
        if val > threshold:
            secMeanOverThresh.append(val)
            indice.append(meanInd[i])
    # now return the data based on the max value and the min value
    maxInd = indice[np.argmax(secMeanOverThresh)]
    minInd = indice[np.argmin(secMeanOverThresh)]

    return data[maxInd:minInd]


def calc_periods_for_files(directory, filenames, forkIsSplit):
    '''Calculates the period for all filenames in directory.

    Parameters
    ----------
    directory : string
        This is the path to the RawData directory.
    filenames : list
        List of all the mat file names in the RawData directory.
    forkIsSplit : boolean
        True if the fork is broken into a handlebar and fork and false if the
        fork and handlebar was measured together.

    Returns
    -------
    periods : dictionary
        Contains all the periods for the mat files in the RawData directory.

    '''

    periods = {}

    def pathParts(path):
        '''Splits a path into a list of its parts.'''
        components = []
        while True:
            (path, tail) = os.path.split(path)
            if tail == "":
                components.reverse()
                return components
            components.append(tail)

    pathToRawDataParts = pathParts(directory)
    pathToRawDataParts.pop()
    pathToBicycleDir = os.path.join(pathToRawDataParts[0],
                                    pathToRawDataParts[1],
                                    pathToRawDataParts[2])
    pathToPlotDir = os.path.join(pathToBicycleDir, 'Plots', 'PendulumFit')

    # make sure there is a place to save the plots
    if not os.path.exists(pathToPlotDir):
        os.makedirs(pathToPlotDir)

    for f in filenames:
        print("Calculating the period for:", f)
        # load the pendulum data
        pathToMatFile = os.path.join(directory, f)
        matData = load_pendulum_mat_file(pathToMatFile)
        # generate a variable name for this period
        periodKey = get_period_key(matData, forkIsSplit)
        # calculate the period
        sampleRate = get_sample_rate(matData)
        pathToPlotFile = os.path.join(pathToPlotDir,
                                      os.path.splitext(f)[0] + '.png')
        period = get_period_from_truncated(matData['data'],
                                           sampleRate,
                                           pathToPlotFile)
        print("The period is:", period, "\n")
        # either append the the period or if it isn't there yet, then
        # make a new list
        try:
            periods[periodKey].append(period)
        except KeyError:
            periods[periodKey] = [period]

    # now average all the periods
    for k, v in periods.items():
        if k.startswith('T'):
            periods[k] = np.mean(v)

    return periods


def check_for_period(mp, forkIsSplit):
    '''Returns whether the fork is split into two pieces and whether the period
    calculations need to happen again.

    Parameters
    ----------
    mp : dictionary
        Dictionary the measured parameters.
    forkIsSplit : boolean
        True if the fork is broken into a handlebar and fork and false if the
        fork and handlebar was measured together.

    Returns
    -------
    forcePeriodCalc : boolean
        True if there wasn't enough period data in mp, false if there was.
    forkIsSplit : boolean
        True if the fork is broken into a handlebar and fork and false if the
        fork and handlebar was measured together.

    '''
    forcePeriodCalc = False
    # Check to see if mp contains at enough periods to not need
    # recalculation
    ncTSum = 0
    ntTSum = 0
    for key in mp.keys():
        # check for any periods in the keys
        if key[:2] == 'Tc':
            ncTSum += 1
        elif key[:2] == 'Tt':
            ntTSum += 1

    # if there isn't enough data then force the period cals again
    if forkIsSplit:
        if ncTSum < 5 or ntTSum < 11:
            forcePeriodCalc = True
    else:
        if ncTSum < 4 or ntTSum < 8:
            forcePeriodCalc = True

    return forcePeriodCalc


def fit_goodness(ym, yp):
    '''
    Calculate the goodness of fit.

    Parameters
    ----------
    ym : ndarray, shape(n,)
        The vector of measured values.
    yp : ndarry, shape(n,)
        The vector of predicted values.

    Returns
    -------
    rsq : float
        The r squared value of the fit.
    SSE : float
        The error sum of squares.
    SST : float
        The total sum of squares.
    SSR : float
        The regression sum of squares.

    '''
    SSR = np.sum((yp - np.mean(ym))**2)
    SST = np.sum((ym - np.mean(ym))**2)
    SSE = SST - SSR
    rsq = SSR / SST
    return rsq, SSE, SST, SSR


def get_period(data, sampleRate, pathToPlotFile):
    '''Returns the period and uncertainty for data resembling a decaying
    oscillation.

    Parameters
    ----------
    data : ndarray, shape(n,)
        A time series that resembles a decaying oscillation.
    sampleRate : int
        The frequency that data was sampled at.
    pathToPlotFile : string
        A path to the file to print the plots.

    Returns
    -------
    T : ufloat
        The period of oscillation and its uncertainty.

    '''

    y = data
    x = np.linspace(0., (len(y) - 1) / float(sampleRate), num=len(y))

    def fitfunc(p, t):
        '''Decaying oscillation function.'''
        a = p[0]
        b = np.exp(-p[3] * p[4] * t)
        c = p[1] * np.sin(p[4] * np.sqrt(1 - p[3]**2) * t)
        d = p[2] * np.cos(p[4] * np.sqrt(1 - p[3]**2) * t)
        return a + b * (c + d)

    # initial guesses
    # p0 = np.array([1.35, -.5, -.75, 0.01, 3.93]) # guess from delft
    # p0 = np.array([2.5, -.75, -.75, 0.001, 4.3]) # guess from ucd
    p0 = make_guess(data, sampleRate)  # tries to make a good guess

    # create the error function
    errfunc = lambda p, t, y: fitfunc(p, t) - y

    # minimize the error function
    p1, success = leastsq(errfunc, p0[:], args=(x, y))

    lscurve = fitfunc(p1, x)

    # find the uncertainty in the fit parameters
    rsq, SSE, SST, SSR = fit_goodness(y, lscurve)
    sigma = np.sqrt(SSE / (len(y) - len(p0)))

    # calculate the jacobian
    L = jac_fitfunc(p1, x)

    # the Hessian
    H = np.dot(L.T, L)

    # the covariance matrix
    U = sigma**2. * np.linalg.inv(H)

    # the standard deviations
    sigp = np.sqrt(U.diagonal())

    # natural frequency
    wo = ufloat(p1[4], sigp[4])
    # damping ratio
    zeta = ufloat(p1[3], sigp[3])
    # damped natural frequency
    wd = (1. - zeta**2.)**(1. / 2.) * wo
    # damped natural frequency (hz)
    fd = wd / 2. / pi
    # period
    T = 1. / fd

    # plot the data and save it to file
    fig = plt.figure()
    plot_osfit(x, y, lscurve, p1, rsq, T, m=np.max(x), fig=fig)
    plt.savefig(pathToPlotFile)
    plt.close()

    # return the period
    return T


def get_period_from_truncated(data, sampleRate, pathToPlotFile):
    # dataRec = average_rectified_sections(data)
    dataRec = data
    dataGood = select_good_data(dataRec, 0.1)
    return get_period(dataGood, sampleRate, pathToPlotFile)


def get_period_key(matData, forkIsSplit):
    '''Returns a dictionary key for the period entries.

    Parameters
    ----------
    matData : dictionary
        The data imported from a pendulum mat file.
    forkIsSplit : boolean
        True if the fork is broken into a handlebar and fork and false if the
        fork and handlebar was measured together.

    Returns
    -------
    key : string
        A key of the form 'T[pendulum][part][orientation]'. For example, if it
        is the frame that was hung as a torsional pendulum at the second
        orientation angle then the key would be 'TtB2'.

    '''
    # set up the subscripting for the period key
    subscripts = {'Fwheel': 'F',
                  'Rwheel': 'R',
                  'Frame': 'B',
                  'Flywheel': 'D'}
    # the Flywheel is for the gyro bike and it actually represents the front
    # wheel and the flywheel as one rigid body. It was easier to measure the
    # the inertia this way. So...the to get the actual flywheel inertia, one
    # must subtract the inertia of the Fwheel, F, from the Flywheel, D.
    if forkIsSplit:
        subscripts['Fork'] = 'S'
        subscripts['Handlebar'] = 'G'
    else:
        subscripts['Fork'] = 'H'
    try:
        subscripts[matData['rod']] = 'P'
    except KeyError:
        subscripts['Rod'] = 'P'

    # used to convert word ordinals to numbers
    ordinal = {'First': '1',
               'Second': '2',
               'Third': '3',
               'Fourth': '4',
               'Fifth': '5',
               'Sixth': '6'}
    try:
        orienWord = matData['angleOrder']
    except:
        orienWord = matData['angle']
    pend = matData['pendulum'][0].lower()
    part = subscripts[matData['part']]
    orienNum = ordinal[orienWord]
    return 'T' + pend + part + orienNum


def get_sample_rate(matData):
    '''Returns the sample rate for the data.'''
    if 'ActualRate' in matData.keys():
        sampleRate = matData['ActualRate']
    else:
        sampleRate = matData['sampleRate']
    return sampleRate


def jac_fitfunc(p, t):
    '''
    Calculate the Jacobian of a decaying oscillation function.

    Uses the analytical formulations of the partial derivatives.

    Parameters
    ----------
    p : the five parameters of the equation
    t : time vector

    Returns
    -------
    jac : The jacobian, the partial of the vector function with respect to the
    parameters vector. A 5 x N matrix where N is the number of time steps.

    '''
    jac = np.zeros((len(p), len(t)))
    e = np.exp(-p[3] * p[4] * t)
    dampsq = np.sqrt(1 - p[3]**2)
    s = np.sin(dampsq * p[4] * t)
    c = np.cos(dampsq * p[4] * t)
    jac[0] = np.ones_like(t)
    jac[1] = e * s
    jac[2] = e * c
    jac[3] = (-p[4] * t * e * (p[1] * s + p[2] * c) + e * (-p[1] * p[3] * p[4]
              * t / dampsq * c + p[2] * p[3] * p[4] * t / dampsq * s))
    jac[4] = (-p[3] * t * e * (p[1] * s + p[2] * c) + e * dampsq * t * (p[1] *
              c - p[2] * s))
    return jac.T


def make_guess(data, sampleRate):
    '''Returns a decent starting point for fitting the decaying oscillation
    function.

    '''
    p = np.zeros(5)

    # the first unknown is the shift along the y axis
    p[0] = np.mean(data)

    # work with the mean subtracted data from now on
    data = data - p[0]

    # what is the initial slope of the curve
    if data[10] > data[0]:
        slope = 1
    else:
        slope = -1

    # the second is the amplitude for the sin function
    p[1] = slope * np.max(data) / 2

    # the third is the amplitude for the cos function
    p[2] = slope * np.max(data)

    # the fourth is the damping ratio and is typically small, 0.001 < zeta <
    # 0.02
    p[3] = 0.001

    # the fifth is the undamped natural frequency
    # first remove the data around zero
    dataMasked = ma.masked_inside(data, -0.1, 0.1)
    # find the zero crossings
    zeroCrossings = np.where(np.diff(np.sign(dataMasked)))[0]
    # remove redundant crossings
    zero = []
    for i, v in enumerate(zeroCrossings):
        if abs(v - zeroCrossings[i - 1]) > 20:
            zero.append(v)
    # get the samples per period
    samplesPerPeriod = 2*np.mean(np.diff(zero))
    # now the frequency
    p[4] = (samplesPerPeriod / float(sampleRate) / 2. / pi)**-1
    if np.isnan(p[4]):
        p[4] = 4.

    return p


def plot_osfit(t, ym, yf, p, rsq, T, m=None, fig=None):
    '''Plot fitted data over the measured

    Parameters
    ----------
    t : ndarray (n,)
        Measurement time in seconds
    ym : ndarray (n,)
        The measured voltage
    yf : ndarray (n,)
    p : ndarray (5,)
        The fit parameters for the decaying osicallation fucntion
    rsq : float
        The r squared value of y (the fit)
    T : float
        The period
    m : float
        The maximum value to plot

    Returns
    -------
    fig : the figure

    '''
    # figure properties
    figwidth = 4.  # in inches
    goldenMean = (np.sqrt(5) - 1.0) / 2.0
    figsize = [figwidth, figwidth * goldenMean]
    params = {#'backend': 'ps',
        'axes.labelsize': 8,
        'axes.titlesize': 8,
        'text.fontsize': 8,
        'legend.fontsize': 8,
        'xtick.labelsize': 6,
        'ytick.labelsize': 6,
        'text.usetex': True,
        #'figure.figsize': figsize
        }
    if fig:
        fig = fig
    else:
        fig = plt.figure(2)
    fig.set_size_inches(figsize)
    plt.rcParams.update(params)
    ax1 = plt.axes([0.125, 0.125, 0.9-0.125, 0.65])
    #if m == None:
        #end = len(t)
    #else:
        #end = t[round(m/t[-1]*len(t))]
    ax1.plot(t, ym, '.', markersize=2)
    plt.plot(t, yf, 'k-')
    plt.xlabel('Time [s]')
    plt.ylabel('Amplitude [V]')
    equation = r'$f(t)={0:1.2f}+e^{{-({3:1.3f})({4:1.1f})t}}\left[{1:1.2f}\sin{{\sqrt{{1-{3:1.3f}^2}}{4:1.1f}t}}+{2:1.2f}\cos{{\sqrt{{1-{3:1.3f}^2}}{4:1.1f}t}}\right]$'.format(p[0], p[1], p[2], p[3], p[4])
    rsquare = '$r^2={0:1.3f}$'.format(rsq)
    period = '$T={0} s$'.format(T)
    plt.title(equation + '\n' + rsquare + ', ' + period)
    plt.legend(['Measured', 'Fit'])
    if m is not None:
        plt.xlim((0, m))
    else:
        pass
    return fig


def select_good_data(data, percent):

    '''Returns a slice of the data from the index at maximum value to the index
    at a percent of the maximum value.

    Parameters
    ----------
    data : ndarray, shape(1,)
        This should be a decaying function.
    percent : float
        The percent of the maximum to clip.

    This basically snips of the beginning and end of the data so that the super
    damped tails are gone and also any weirdness at the beginning.

    '''
    meanSub = data - np.mean(data)
    maxVal = np.max(np.abs(meanSub))
    maxInd = np.argmax(np.abs(meanSub))
    for i, v in reversed(list(enumerate(meanSub))):
        if v > percent * maxVal:
            minInd = i
            break

    return data[maxInd:minInd]



#!/usr/bin/env python

import os
import numpy as np
from numpy import sin, cos, sqrt
from scipy.optimize import fsolve
import yeadon

from .io import remove_uncertainties
from .inertia import combine_bike_rider


def yeadon_vec_to_bicycle_vec(vector, measured_bicycle_par,
                              benchmark_bicycle_par):
    """
    Parameters
    ----------
    vector : ndarray, shape(3, 1)
        A vector from the Yeadon origin to a point expressed in the Yeadon
        reference frame.
    measured_bicycle_par : dictionary
        The raw bicycle measurements.
    benchmark_bicycle_par : dictionary
        The Meijaard 2007 et. al parameters for this bicycle.

    Returns
    -------
    vector_wrt_bike : ndarray, shape(3, 1)
        The vector from the bicycle origin to the same point above expressed
        in the bicycle reference frame.

    """

    # This is the rotation matrix that relates Yeadon's reference frame
    # to the bicycle reference frame.
    # vector_expressed_in_bike = rot_mat * vector_expressed_in_yeadon)
    rot_mat = np.array([[0.0, -1.0, 0.0],
                        [-1.0, 0.0, 0.0],
                        [0.0, 0.0, -1.0]])

    # The relevant bicycle parameters:
    measuredPar = remove_uncertainties(measured_bicycle_par)
    benchmarkPar = remove_uncertainties(benchmark_bicycle_par)
    # bottom bracket height
    hbb = measuredPar['hbb']
    # chain stay length
    lcs = measuredPar['lcs']
    # rear wheel radius
    rR = benchmarkPar['rR']
    # seat post length
    lsp = measuredPar['lsp']
    # seat tube length
    lst = measuredPar['lst']
    # seat tube angle
    lambdast = measuredPar['lamst']

    # bicycle origin to yeadon origin expressed in bicycle frame
    yeadon_origin_in_bike_frame = \
        np.array([[np.sqrt(lcs**2 - (-hbb + rR)**2) +
                   (-lsp - lst) * np.cos(lambdast)],  # bx
                  [0.0],
                  [-hbb + (-lsp - lst) * np.sin(lambdast)]])  # bz

    vector_wrt_bike = yeadon_origin_in_bike_frame + rot_mat @ vector

    return vector_wrt_bike


def configure_rider(pathToRider, bicycle, bicyclePar, measuredPar, draw):
    """
    Returns the rider parameters, bicycle paramaters with a rider and a
    human object that is configured to sit on the bicycle.

    Parameters
    ----------
    pathToRider : string
        Path to the rider's data folder.
    bicycle : string
        The short name of the bicycle.
    bicyclePar : dictionary
        Contains the benchmark bicycle parameters for a bicycle.
    measuredPar : dictionary
        Contains the measured values of the bicycle.
    draw : boolean, optional
        If true, visual python will be used to draw a three dimensional
        image of the rider.

    Returns
    -------
    riderpar : dictionary
        The inertial parameters of the rider with reference to the
        benchmark coordinate system.
    human : yeadon.human
        The human object that represents the rider seated on the
        bicycle.
    bicycleRiderPar : dictionary
        The benchmark parameters of the bicycle with the rider added to
        the rear frame.

    """
    try:
        # get the rider name
        rider = os.path.split(pathToRider)[1]
        # get the paths to the yeadon data files
        pathToYeadon = os.path.join(pathToRider, 'RawData',
                                    rider + 'YeadonMeas.txt')
        pathToCFG = os.path.join(pathToRider, 'RawData',
                                 rider + bicycle + 'YeadonCFG.txt')
        # generate the human that has been configured to sit on the bicycle
        # the human's inertial parameters are expressed in the Yeadon
        # reference frame about the Yeadon origin.
        human = rider_on_bike(bicyclePar, measuredPar,
                              pathToYeadon, pathToCFG, draw)

        # This is the rotation matrix that relates Yeadon's reference frame
        # to the bicycle reference frame.
        rot_mat = np.array([[0.0, -1.0, 0.0],
                            [-1.0, 0.0, 0.0],
                            [0.0, 0.0, -1.0]])

        # This is the human's inertia expressed in the bicycle reference
        # frame about the human's center of mass.
        human_inertia_in_bike_frame = \
            human.inertia_transformed(rotmat=rot_mat)

        human_com_in_bike_frame = \
            yeadon_vec_to_bicycle_vec(human.center_of_mass, measuredPar,
                                      bicyclePar)

        # build a dictionary to store the inertial data
        riderPar = {'IBxx': human_inertia_in_bike_frame[0, 0],
                    'IByy': human_inertia_in_bike_frame[1, 1],
                    'IBzz': human_inertia_in_bike_frame[2, 2],
                    'IBxz': human_inertia_in_bike_frame[2, 0],
                    'mB': human.mass,
                    'xB': human_com_in_bike_frame[0, 0],
                    'yB': human_com_in_bike_frame[1, 0],
                    'zB': human_com_in_bike_frame[2, 0]}
    except:  # except if this fails
        # no rider was added
        print('Calculations in yeadon failed. No rider added.')
        # raise the error that caused things to fail
        raise
    else:
        bicycleRiderPar = combine_bike_rider(bicyclePar, riderPar)
        return riderPar, human, bicycleRiderPar


def rider_on_bike(benchmarkPar, measuredPar, yeadonMeas, yeadonCFG,
                  drawrider):
    """
    Returns a yeadon human configured to sit on a bicycle.

    Parameters
    ----------
    benchmarkPar : dictionary
        A dictionary containing the benchmark bicycle parameters.
    measuredPar : dictionary
        A dictionary containing the raw geometric measurements of the bicycle.
    yeadonMeas : str
        Path to a text file that holds the 95 yeadon measurements. See
        `yeadon documentation`_.
    yeadonCFG : str
        Path to a text file that holds configuration variables. See `yeadon
        documentation`_. As of now, only 'somersalt' angle can be set as an
        input. The remaining variables are either zero or calculated in this
        method.
    drawrider : bool
        Switch to draw the rider, with vectors pointing to the desired
        position of the hands and feet of the rider (at the handles and
        bottom bracket). Requires python-visual.

    Returns
    -------
    human : yeadon.Human
        Human object is returned with an updated configuration.
        The dictionary, taken from H.CFG, has the following key's values
        updated:

            'PJ1extension'
            'J1J2flexion'
            'CA1extension'
            'CA1adduction'
            'CA1rotation'
            'A1A2extension'
            'somersault'
            'PK1extension'
            'K1K2flexion'
            'CB1extension'
            'CB1abduction'
            'CB1rotation'
            'B1B2extension'

    Notes
    -----
    Requires that the bike object has a raw data text input file that contains
    the measurements necessary to situate a rider on the bike (i.e.
    ``<pathToData>/bicycles/<short name>/RawData/<short name>Measurements.txt``).

    .. _yeadon documentation : http://packages.python.org/yeadon


    """

    # create human using input measurements and configuration files
    human = yeadon.Human(yeadonMeas, yeadonCFG)

    # The relevant human measurments:
    L_j3L = human.meas['Lj3L']
    L_j5L = human.meas['Lj5L']
    L_j6L = human.meas['Lj6L']
    L_s4L = human.meas['Ls4L']
    L_s4w = human.meas['Ls4w']
    L_a2L = human.meas['La2L']
    L_a4L = human.meas['La4L']
    L_a5L = human.meas['La5L']
    somersault = human.CFG['somersault']

    # The relevant bicycle parameters:
    measuredPar = remove_uncertainties(measuredPar)
    benchmarkPar = remove_uncertainties(benchmarkPar)
    # bottom bracket height
    h_bb = measuredPar['hbb']
    # chain stay length
    l_cs = measuredPar['lcs']
    # rear wheel radius
    r_R = benchmarkPar['rR']
    # front wheel radius
    r_F = benchmarkPar['rF']
    # seat post length
    l_sp = measuredPar['lsp']
    # seat tube length
    l_st = measuredPar['lst']
    # seat tube angle
    lambda_st = measuredPar['lamst']
    # handlebar width
    w_hb = measuredPar['whb']
    # distance from rear wheel hub to hand
    L_hbR = measuredPar['LhbR']
    # distance from front wheel hub to hand
    L_hbF = measuredPar['LhbF']
    # wheelbase
    w = benchmarkPar['w']

    def zero(unknowns):
        """For the derivation of these equations see:

           http://nbviewer.ipython.org/github/chrisdembia/yeadon/blob/v1.2.0/examples/bicyclerider/bicycle_example.ipynb
        """

        PJ1extension = unknowns[0]
        J1J2flexion = unknowns[1]
        CA1extension = unknowns[2]
        CA1adduction = unknowns[3]
        CA1rotation = unknowns[4]
        A1A2extension = unknowns[5]
        alpha_y = unknowns[6]
        alpha_z = unknowns[7]
        beta_y = unknowns[8]
        beta_z = unknowns[9]

        phi_J1 = PJ1extension
        phi_J2 = J1J2flexion
        phi_A1 = CA1extension
        theta_A1 = CA1adduction
        psi_A = CA1rotation
        phi_A2 = A1A2extension

        phi_P = somersault

        zero = np.zeros(10)

        zero[0] = (L_j3L*(-sin(phi_J1)*cos(phi_P) - sin(phi_P)*cos(phi_J1))
                   + (-l_sp - l_st)*cos(lambda_st) + (-(-sin(phi_J1)*
                   sin(phi_P) + cos(phi_J1)*cos(phi_P))*sin(phi_J2) +
                   (-sin(phi_J1)*cos(phi_P) - sin(phi_P)*cos(phi_J1))*
                   cos(phi_J2))*(-L_j3L + L_j5L + L_j6L))

        zero[1] = (L_j3L*(-sin(phi_J1)*sin(phi_P) + cos(phi_J1)*cos(phi_P))
                   + (-l_sp - l_st)*sin(lambda_st) + ((-sin(phi_J1)*
                   sin(phi_P) + cos(phi_J1)*cos(phi_P))*cos(phi_J2) -
                   (sin(phi_J1)*cos(phi_P) + sin(phi_P)*cos(phi_J1))*
                   sin(phi_J2))*(-L_j3L + L_j5L + L_j6L))

        zero[2] = -L_hbF + sqrt(alpha_y**2 + alpha_z**2 + 0.25*w_hb**2)

        zero[3] = -L_hbR + sqrt(beta_y**2 + beta_z**2 + 0.25*w_hb**2)

        zero[4] = alpha_y - beta_y - w

        zero[5] = alpha_z - beta_z + r_F - r_R

        zero[6] = (-L_a2L*sin(theta_A1) + L_s4w/2 - 0.5*w_hb + (sin(phi_A2)*
                   sin(psi_A)*cos(theta_A1) + sin(theta_A1)*cos(phi_A2))*
                   (L_a2L - L_a4L - L_a5L))

        zero[7] = (-L_a2L*(-sin(phi_A1)*cos(phi_P)*cos(theta_A1) -
                   sin(phi_P)*cos(phi_A1)*cos(theta_A1)) - L_s4L*sin(phi_P)
                   - beta_y - sqrt(l_cs**2 - (-h_bb + r_R)**2) - (-l_sp -
                   l_st)*cos(lambda_st) + (-(-(sin(phi_A1)*cos(psi_A) +
                   sin(psi_A)*sin(theta_A1)*cos(phi_A1))*sin(phi_P) +
                   (-sin(phi_A1)*sin(psi_A)*sin(theta_A1) + cos(phi_A1)*
                   cos(psi_A))*cos(phi_P))*sin(phi_A2) + (-sin(phi_A1)*
                   cos(phi_P)*cos(theta_A1) - sin(phi_P)*cos(phi_A1)*
                   cos(theta_A1))*cos(phi_A2))*(L_a2L - L_a4L - L_a5L))

        zero[8] = (-L_a2L*(-sin(phi_A1)*sin(phi_P)*cos(theta_A1) +
                   cos(phi_A1)*cos(phi_P)*cos(theta_A1)) + L_s4L*cos(phi_P)
                   - beta_z + h_bb - r_R - (-l_sp - l_st)*sin(lambda_st) +
                   (-((sin(phi_A1)*cos(psi_A) + sin(psi_A)*sin(theta_A1)*
                   cos(phi_A1))*cos(phi_P) + (-sin(phi_A1)*sin(psi_A)*
                   sin(theta_A1) + cos(phi_A1)*cos(psi_A))*sin(phi_P))*
                   sin(phi_A2) + (-sin(phi_A1)*sin(phi_P)*cos(theta_A1) +
                   cos(phi_A1)*cos(phi_P)*cos(theta_A1))*cos(phi_A2))*(L_a2L
                   - L_a4L - L_a5L))

        zero[9] = ((sin(phi_A1)*sin(psi_A) - sin(theta_A1)*cos(phi_A1)*
                    cos(psi_A))*cos(phi_P) + (sin(phi_A1)*sin(theta_A1)*
                    cos(psi_A) + sin(psi_A)*cos(phi_A1))*sin(phi_P))

        return zero

    g_PJ1extension = -np.deg2rad(90.0)
    g_J1J2flexion = np.deg2rad(75.0)
    g_CA1extension = -np.deg2rad(15.0)
    g_CA1adduction = np.deg2rad(2.0)
    g_CA1rotation = np.deg2rad(2.0)
    g_A1A2extension = -np.deg2rad(40.0)
    g_alpha_y = L_hbF * np.cos(np.deg2rad(45.0))
    g_alpha_z = L_hbF * np.sin(np.deg2rad(45.0))
    g_beta_y = -L_hbR * np.cos(np.deg2rad(30.0))
    g_beta_z = L_hbR * np.sin(np.deg2rad(30.0))

    guess = [g_PJ1extension, g_J1J2flexion, g_CA1extension, g_CA1adduction,
             g_CA1rotation, g_A1A2extension, g_alpha_y, g_alpha_z, g_beta_y,
             g_beta_z]

    solution = fsolve(zero, guess)

    cfg_dict = human.CFG.copy()
    cfg_dict['PJ1extension'] = solution[0]
    cfg_dict['J1J2flexion'] = solution[1]
    cfg_dict['CA1extension'] = solution[2]
    cfg_dict['CA1adduction'] = solution[3]
    cfg_dict['CA1rotation'] = solution[4]
    cfg_dict['A1A2extension'] = solution[5]
    cfg_dict['somersault'] = somersault
    cfg_dict['PK1extension'] = cfg_dict['PJ1extension']
    cfg_dict['K1K2flexion'] = cfg_dict['J1J2flexion']
    cfg_dict['CB1extension'] = cfg_dict['CA1extension']
    cfg_dict['CB1abduction'] = -cfg_dict['CA1adduction']
    cfg_dict['CB1rotation'] = -cfg_dict['CA1rotation']
    cfg_dict['B1B2extension'] = cfg_dict['A1A2extension']

    # assign configuration to human and check that the solution worked
    human.set_CFG_dict(cfg_dict)

    # draw rider for fun, but possibly to check results aren't crazy
    if drawrider:
        human.draw()

    return human



# import all the important classes and functions from the bicycleparametes
# module
from .main import Bicycle
from .plot import plot_eigenvalues, compare_bode_bicycles
from .tables import Table
from .version import __version_info__, __version__

# the modules that are imported when 'from bicycleparameters import *'
__all__ = [
    'bicycle',
    'com',
    'geometry',
    'inertia',
    'io',
    'main',
    'period',
    'plot',
    'rider',
    'tables',
]



#!/usr/bin/env python

import numpy as np
from scipy.optimize import newton
from uncertainties import umath, ufloat


def ab_matrix(M, C1, K0, K2, v, g):
    '''Calculate the A and B matrices for the Whipple bicycle model linearized
    about the upright configuration.

    Parameters
    ----------
    M : ndarray, shape(2,2)
        The mass matrix.
    C1 : ndarray, shape(2,2)
        The damping like matrix that is proportional to the speed, v.
    K0 : ndarray, shape(2,2)
        The stiffness matrix proportional to gravity, g.
    K2 : ndarray, shape(2,2)
        The stiffness matrix proportional to the speed squared, v**2.
    v : float
        Forward speed.
    g : float
        Acceleration due to gravity.

    Returns
    -------
    A : ndarray, shape(4,4)
        State matrix.
    B : ndarray, shape(4,2)
        Input matrix.

    The states are [roll angle,
                    steer angle,
                    roll rate,
                    steer rate]
    The inputs are [roll torque,
                    steer torque]

    '''

    invM = (1. / (M[0, 0] * M[1, 1] - M[0, 1] * M[1, 0]) *
            np.array([[M[1, 1], -M[0, 1]], [-M[1, 0], M[0, 0]]],
                     dtype=M.dtype))

    a11 = np.zeros((2, 2))
    a12 = np.eye(2)
    # stiffness based terms
    a21 = -np.dot(invM, (g * K0 + v**2 * K2))
    # damping based terms
    a22 = -np.dot(invM, v * C1)

    A = np.vstack((np.hstack((a11, a12)), np.hstack((a21, a22))))
    B = np.vstack((np.zeros((2, 2)), invM))

    return A, B


def benchmark_par_to_canonical(p):
    '''Returns the canonical matrices of the Whipple bicycle model linearized
    about the upright constant velocity configuration. It uses the parameter
    definitions from Meijaard et al. 2007.

    Parameters
    ----------
    p : dictionary
        A dictionary of the benchmark bicycle parameters. Make sure your units
        are correct, best to ue the benchmark paper's units!

    Returns
    -------
    M : ndarray, shape(2,2)
        The mass matrix.
    C1 : ndarray, shape(2,2)
        The damping like matrix that is proportional to the speed, v.
    K0 : ndarray, shape(2,2)
        The stiffness matrix proportional to gravity, g.
    K2 : ndarray, shape(2,2)
        The stiffness matrix proportional to the speed squared, v**2.

    Notes
    -----
    This function handles parameters with uncertanties.

    '''
    mT = p['mR'] + p['mB'] + p['mH'] + p['mF']
    xT = (p['xB']*p['mB'] + p['xH']*p['mH'] + p['w']*p['mF']) / mT
    zT = (-p['rR']*p['mR'] + p['zB']*p['mB'] +
          p['zH']*p['mH'] - p['rF']*p['mF']) / mT

    ITxx = (p['IRxx'] + p['IBxx'] + p['IHxx'] + p['IFxx'] + p['mR'] *
            p['rR']**2 + p['mB']*p['zB']**2 + p['mH']*p['zH']**2 +
            p['mF']*p['rF']**2)
    ITxz = (p['IBxz'] + p['IHxz'] - p['mB']*p['xB']*p['zB'] -
            p['mH']*p['xH']*p['zH'] + p['mF']*p['w']*p['rF'])
    p['IRzz'] = p['IRxx']
    p['IFzz'] = p['IFxx']
    ITzz = (p['IRzz'] + p['IBzz'] + p['IHzz'] + p['IFzz'] +
            p['mB']*p['xB']**2 + p['mH']*p['xH']**2 + p['mF']*p['w']**2)

    mA = p['mH'] + p['mF']
    xA = (p['xH']*p['mH'] + p['w']*p['mF']) / mA
    zA = (p['zH']*p['mH'] - p['rF']*p['mF']) / mA

    IAxx = (p['IHxx'] + p['IFxx'] + p['mH']*(p['zH'] - zA)**2 +
            p['mF']*(p['rF'] + zA)**2)
    IAxz = (p['IHxz'] - p['mH']*(p['xH'] - xA)*(p['zH'] - zA) + p['mF'] *
            (p['w'] - xA)*(p['rF'] + zA))
    IAzz = (p['IHzz'] + p['IFzz'] + p['mH']*(p['xH'] - xA)**2 + p['mF'] *
            (p['w'] - xA)**2)
    uA = (xA - p['w'] - p['c'])*umath.cos(p['lam']) - zA*umath.sin(p['lam'])
    IAll = (mA*uA**2 + IAxx*umath.sin(p['lam'])**2 +
            2*IAxz*umath.sin(p['lam'])*umath.cos(p['lam']) +
            IAzz*umath.cos(p['lam'])**2)
    IAlx = (-mA*uA*zA + IAxx*umath.sin(p['lam']) + IAxz*umath.cos(p['lam']))
    IAlz = (mA*uA*xA + IAxz*umath.sin(p['lam']) + IAzz*umath.cos(p['lam']))

    mu = p['c'] / p['w']*umath.cos(p['lam'])

    SR = p['IRyy'] / p['rR']
    SF = p['IFyy'] / p['rF']
    ST = SR + SF
    SA = mA*uA + mu*mT*xT

    Mpp = ITxx
    Mpd = IAlx + mu*ITxz
    Mdp = Mpd
    Mdd = IAll + 2*mu*IAlz + mu**2*ITzz
    M = np.array([[Mpp, Mpd], [Mdp, Mdd]])

    K0pp = mT*zT  # this value only reports to 13 digit precision it seems?
    K0pd = -SA
    K0dp = K0pd
    K0dd = -SA*umath.sin(p['lam'])
    K0 = np.array([[K0pp, K0pd], [K0dp, K0dd]])

    K2pp = 0.
    K2pd = (ST - mT*zT) / p['w']*umath.cos(p['lam'])
    K2dp = 0.
    K2dd = (SA + SF*umath.sin(p['lam'])) / p['w']*umath.cos(p['lam'])
    K2 = np.array([[K2pp, K2pd], [K2dp, K2dd]])

    C1pp = 0.
    C1pd = (mu*ST + SF*umath.cos(p['lam']) + ITxz / p['w'] *
            umath.cos(p['lam']) - mu*mT*zT)
    C1dp = -(mu*ST + SF*umath.cos(p['lam']))
    C1dd = (IAlz / p['w']*umath.cos(p['lam']) + mu*(SA +
            ITzz / p['w']*umath.cos(p['lam'])))
    C1 = np.array([[C1pp, C1pd], [C1dp, C1dd]])

    return M, C1, K0, K2


def lambda_from_abc(rF, rR, a, b, c):
    '''Returns the steer axis tilt, lamba, for the parameter set based on the
    offsets from the steer axis.

    Parameters
    ----------
    rF : float or ufloat
        Front wheel radius.
    rR : float or ufloat
        Rear wheel radius.
    a : float or ufloat
        The rear wheel offset. The minimum distance from the steer axis to the
        center of the rear wheel.
    b : float or ufloat
        The front wheel offset. The minimum distance from the steer axis to the
        center of the front wheel.
    c : float or ufloat
        The steer axis distance. The distance along the steer axis between the
        intersection of the front and rear wheel offset lines.

    '''
    def lam_equality(lam, rF, rR, a, b, c):
        return umath.sin(lam) - (rF - rR + c * umath.cos(lam)) / (a + b)
    guess = umath.atan(c / (a + b))  # guess based on equal wheel radii

    # The following assumes that the uncertainty calculated for the guess is
    # the same as the uncertainty for the true solution. This is not true! and
    # will surely breakdown the further the guess is away from the true
    # solution. There may be a way to calculate the correct uncertainity, but
    # that needs to be figured out. I guess I could use least squares and do it
    # the same way as get_period.

    args = (rF.nominal_value, rR.nominal_value, a.nominal_value,
            b.nominal_value, c.nominal_value)

    lam = newton(lam_equality, guess.nominal_value, args=args)
    return ufloat(lam, guess.std_dev)


def sort_eigenmodes(evals, evecs):
    """Sort eigenvalues and eigenvectors.

    Parameters
    ==========
    evals : ndarray, shape (n, 4)
        A sequence of n sets of eigenvalues.
    evecs : ndarray, shape (n, 4, 4)
        A sequence of n sets of eigenvectors.

    Returns
    =======
    evalsorg : ndarray, shape (n, 4)
        A sequence of n sets of eigenvalues.
    evecsorg : ndarray, shape (n, 4, 4)
        A sequence of n sets of eigenvectors.

    """
    evalsorg = np.zeros_like(evals)
    evecsorg = np.zeros_like(evecs)
    # set the first row to be the same
    evalsorg[0] = evals[0]
    evecsorg[0] = evecs[0]
    # for each speed
    for i, speed in enumerate(evals):
        if i == evals.shape[0] - 1:
            break
        # for each current eigenvalue
        used = []
        for j, e in enumerate(speed):
            x, y = np.real(evalsorg[i, j]), np.imag(evalsorg[i, j])
            # for each eigenvalue at the next speed
            dist = np.zeros(4)
            for k, eignext in enumerate(evals[i + 1]):
                xn, yn = np.real(eignext), np.imag(eignext)
                # distance between points in the real/imag plane
                dist[k] = np.abs(((xn - x)**2 + (yn - y)**2)**0.5)
            if np.argmin(dist) in used:
                # set the already used indice higher
                dist[np.argmin(dist)] = np.max(dist) + 1.
            else:
                pass
            evalsorg[i + 1, j] = evals[i + 1, np.argmin(dist)]
            evecsorg[i + 1, :, j] = evecs[i + 1, :, np.argmin(dist)]
            # keep track of the indices we've used
            used.append(np.argmin(dist))

    return evalsorg, evecsorg


def sort_modes(evals, evecs):
    '''Sort eigenvalues and eigenvectors into weave, capsize, caster modes.

    Parameters
    ----------
    evals : ndarray, shape (n, 4)
        eigenvalues
    evecs : ndarray, shape (n, 4, 4)
        eigenvectors

    Returns
    -------
    weave['evals'] : ndarray, shape (n, 2)
        The eigen value pair associated with the weave mode.
    weave['evecs'] : ndarray, shape (n, 4, 2)
        The associated eigenvectors of the weave mode.
    capsize['evals'] : ndarray, shape (n,)
        The real eigenvalue associated with the capsize mode.
    capsize['evecs'] : ndarray, shape(n, 4, 1)
        The associated eigenvectors of the capsize mode.
    caster['evals'] : ndarray, shape (n,)
        The real eigenvalue associated with the caster mode.
    caster['evecs'] : ndarray, shape(n, 4, 1)
        The associated eigenvectors of the caster mode.

    Notes
    -----
    This only works on the standard bicycle eigenvalues, not necessarily on any
    general eigenvalues for the bike model (e.g. there isn't always a distinct
    weave, capsize and caster). Some type of check using the derivative of the
    curves could make it more robust.

    '''
    evalsorg = np.zeros_like(evals)
    evecsorg = np.zeros_like(evecs)
    # set the first row to be the same
    evalsorg[0] = evals[0]
    evecsorg[0] = evecs[0]
    # for each speed
    for i, speed in enumerate(evals):
        if i == evals.shape[0] - 1:
            break
        # for each current eigenvalue
        used = []
        for j, e in enumerate(speed):
            try:
                x = np.real(evalsorg[i, j].nominal_value)
                y = np.imag(evalsorg[i, j].nominal_value)
            except:
                x, y = np.real(evalsorg[i, j]), np.imag(evalsorg[i, j])
            # for each eigenvalue at the next speed
            dist = np.zeros(4)
            for k, eignext in enumerate(evals[i + 1]):
                try:
                    xn = np.real(eignext.nominal_value)
                    yn = np.imag(eignext.nominal_value)
                except:
                    xn, yn = np.real(eignext), np.imag(eignext)
                # distance between points in the real/imag plane
                dist[k] = np.abs(((xn - x)**2 + (yn - y)**2)**0.5)
            if np.argmin(dist) in used:
                # set the already used indice higher
                dist[np.argmin(dist)] = np.max(dist) + 1.
            else:
                pass
            evalsorg[i + 1, j] = evals[i + 1, np.argmin(dist)]
            evecsorg[i + 1, :, j] = evecs[i + 1, :, np.argmin(dist)]
            # keep track of the indices we've used
            used.append(np.argmin(dist))
    weave = {'evals': evalsorg[:, 2:], 'evecs': evecsorg[:, :, 2:]}
    capsize = {'evals': evalsorg[:, 1], 'evecs': evecsorg[:, :, 1]}
    caster = {'evals': evalsorg[:, 0], 'evecs': evecsorg[:, :, 0]}
    return weave, capsize, caster


def trail(rF, lam, fo):
    '''Calculate the trail and mechanical trail.

    Parameters
    ----------
    rF : float
        The front wheel radius
    lam : float
        The steer axis tilt (pi/2 - headtube angle). The angle between the
        headtube and a vertical line.
    fo : float
        The fork offset

    Returns
    -------
    c: float
        Trail
    cm: float
        Mechanical Trail

    '''

    # trail
    c = (rF * umath.sin(lam) - fo) / umath.cos(lam)
    # mechanical trail
    cm = c * umath.cos(lam)
    return c, cm
